/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XMLHelper.java 17:06:54 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.xmlmanipulation;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.OutputStream;
import java.util.List;
import java.util.Map;

import javax.wsdl.BindingOperation;
import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.xml.WSDLReader;
import javax.wsdl.xml.WSDLWriter;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;

import com.ibm.wsdl.xml.WSDLReaderImpl;
import com.ibm.wsdl.xml.WSDLWriterImpl;

/**
 * Util class to manipulate WSDL Document or Definition
 * 
 * @author wjoseph - eBMWebsourcing
 */
public final class WSDLHelper {

    /**
     * Namespace for the soap 1.1
     */
    // private static final String WSDL_NS = "http://schemas.xmlsoap.org/wsdl/";
    private WSDLHelper() {
    }

    /**
     * Change the location attribute of the service node to
     * "/axis2/services/serviceName"
     * 
     * @param serviceName
     *            Name of the service
     * @param serviceDesc
     *            Dom representation of the service description
     */
    public static void changeServiceAddressInWSDL(QName serviceName,
            Definition serviceDef, String prefix) {
        SOAPAddress address = findServiceAddress(serviceDef, serviceName);
        if (address != null) {
            String serviceAddress = prefix + serviceName.getLocalPart();
            address.setLocationURI(serviceAddress);
        }
    }

    /**
     * Change the location attribute of the service node to
     * "/axis2/services/serviceName"
     * 
     * @param serviceName
     *            Name of the service
     * @param serviceDesc
     *            Dom representation of the service description
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static Document changeServiceAddressInWSDL(QName serviceName,
            Document serviceDesc, String prefix) throws WSDLException, TransformerException {
    	String docString = XMLHelper.createStringFromDOMDocument(serviceDesc);
        Document tmpDoc = XMLHelper.createDocumentFromString(docString);
        Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        changeServiceAddressInWSDL(serviceName, serviceDef, prefix);
        serviceDesc = createDocumentFromDefinition(serviceDef);
        return serviceDesc;
    }

    public static Definition createDefinitionFromDocument(Document doc)
        throws WSDLException {
        WSDLReader reader = new WSDLReaderImpl();
        return reader.readWSDL(null, doc);
    }

    /**
     * Create a dom document from a File
     * 
     * @param wsdlFile
     *            file
     * @return dom document
     */
    public static Document createDocumentFromWSDL(File wsdlFile) {
        Document doc = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                    .newInstance();
            factory.setNamespaceAware(true);
            DocumentBuilder builder = factory.newDocumentBuilder();
            doc = builder.parse(wsdlFile);
            doc.normalize();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return doc;
    }

    @SuppressWarnings("unchecked")
    public static SOAPAddress findServiceAddress(Definition serviceDef,
            QName serviceName) {
        SOAPAddress result = null;
        if (serviceDef != null) {
            Service service = serviceDef.getService(serviceName);
            if (service != null) {
                Map<String, Port> ports = service.getPorts();
                for (Port port : ports.values()) {
                    List<ExtensibilityElement> extensibilityElements = port
                            .getExtensibilityElements();
                    for (ExtensibilityElement element : extensibilityElements) {
                        // The SOAP binding is an extensibility element of
                        // Port
                        if (element instanceof SOAPAddress) {
                            result = (SOAPAddress) element;
                            break;
                        }
                    }
                    if (result != null) {
                        break;
                    }
                }
            }
        }
        return result;
    }

    /**
     * Returns the service's port name contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service port name attribute or null if not found
     */
    @SuppressWarnings("unchecked")
    public static String getPortNameFromWSDLDefinition(Definition serviceDef,
            QName serviceName) {
        String result = null;
        Service service = serviceDef.getService(serviceName);
        Map<String, Port> ports = service.getPorts();
        for (Port port : ports.values()) {
            result = port.getName();
        }
        return result;
    }

    /**
     * Returns the service's port name contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service port name attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static String getPortNameFromWSDLDocument(Document serviceDesc,
            QName serviceName) throws WSDLException, TransformerException {
    	String docString = XMLHelper.createStringFromDOMDocument(serviceDesc);
        Document tmpDoc = XMLHelper.createDocumentFromString(docString);
        Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getPortNameFromWSDLDefinition(serviceDef, serviceName);
    }

    /**
     * Returns the service address contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service address attribute or null if not found
     */
    public static String getServiceAddressFromWSDLDefinition(
            Definition serviceDef, QName serviceName) {
        String result = null;
        SOAPAddress address = findServiceAddress(serviceDef, serviceName);
        if (address != null) {
            result = address.getLocationURI();
        }
        return result;
    }

    /**
     * Returns the service address contained in a WSDL 1.0 document
     * 
     * @param serviceDesc
     *            The service description document
     * @return The service address attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static String getServiceAddressFromWSDLDocument(
            Document serviceDesc, QName serviceName) throws WSDLException, TransformerException {
    	String docString = XMLHelper.createStringFromDOMDocument(serviceDesc);
        Document tmpDoc = XMLHelper.createDocumentFromString(docString);
        Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getServiceAddressFromWSDLDefinition(serviceDef, serviceName);
    }

    /**
     * Returns the service address contained in a WSDL 1.0 file
     * 
     * @param wsdlFile
     *            The WSDL file
     * @return The service address attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static String getServiceAddressFromWSDLFile(File wsdlFile,
            QName serviceName) throws WSDLException, TransformerException {
        Document doc = createDocumentFromWSDL(wsdlFile);
        return getServiceAddressFromWSDLDocument(doc, serviceName);
    }

    /**
     * Returns the service name contained in a WSDL 1.1 dom Document
     * 
     * @param serviceDef
     *            The dom Document
     * @return The service name attribute or null if not found
     */
    @SuppressWarnings("unchecked")
    public static QName[] getServiceNameFromWSDLDefinition(Definition serviceDef) {
        QName[] result = null;
        if (serviceDef != null) {
            Map<String, Service> services = serviceDef.getServices();
            result = new QName[services.size()];
            int i = 0;
            for (Service service : services.values()) {
                result[i++] = service.getQName();
            }
        }
        return result;
    }

    /**
     * Returns the service name contained in a WSDL 1.1 dom Document
     * 
     * @param serviceDef
     *            The dom Document
     * @return The service name attribute or null if not found
     * @throws WSDLException 
     * @throws TransformerException 
     */
    public static QName[] getServiceNameFromWSDLDocument(Document serviceDesc)
        throws WSDLException, TransformerException {
        String docString = XMLHelper.createStringFromDOMDocument(serviceDesc);
        Document tmpDoc = XMLHelper.createDocumentFromString(docString);
        Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getServiceNameFromWSDLDefinition(serviceDef);
    }

    /**
     * Returns the service name contained in a WSDL 1.0 file
     * 
     * @param wsdlFile
     *            The WSDL file
     * @return The service name attribute or null if not found
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static QName[] getServiceNameFromWSDLFile(File wsdlFile)
        throws WSDLException, TransformerException {
        Document doc = createDocumentFromWSDL(wsdlFile);
        return getServiceNameFromWSDLDocument(doc);
    }

    /**
     * Returns WSDL 1.0 document target namespace
     * 
     * @param serviceDesc
     *            The service description document
     * @return The target namespace or null if not found
     */
    public static String getTargetNSFromWSDLDefinition(Definition serviceDef) {
        return serviceDef.getTargetNamespace();
    }

    /**
     * Returns WSDL 1.0 document target namespace
     * 
     * @param serviceDesc
     *            The service description document
     * @return The target namespace or null if not found
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static String getTargetNSFromWSDLDocument(Document serviceDesc)
        throws WSDLException, TransformerException {
    	String docString = XMLHelper.createStringFromDOMDocument(serviceDesc);
        Document tmpDoc = XMLHelper.createDocumentFromString(docString);
        Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return getTargetNSFromWSDLDefinition(serviceDef);
    }

    /**
     * Search the service description for an operation
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found or else false
     */
    public static boolean hasOperationNamed(Definition serviceDef,
            String opName, QName serviceName) {
        BindingOperation op = findOperationNamed(serviceDef, opName,
                serviceName);
        return op != null;
    }

    /**
     * Search the service description for an operation
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found or else false
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static boolean hasOperationNamed(Document serviceDesc,
            String opName, QName serviceName) throws WSDLException, TransformerException {
    	String docString = XMLHelper.createStringFromDOMDocument(serviceDesc);
        Document tmpDoc = XMLHelper.createDocumentFromString(docString);
        Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return hasOperationNamed(serviceDef, opName, serviceName);
    }

    /**
     * Search the service description for an operation. If the operation is
     * found , checks if it has a child node named 'output'
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found and has the output child
     *         node or else false
     */
    public static boolean isInOutOperation(Definition serviceDef,
            String opName, QName serviceName) {
        boolean inOut = false;
        BindingOperation op = findOperationNamed(serviceDef, opName,
                serviceName);
        if (op != null) {
            inOut = op.getBindingOutput() != null;
        }
        return inOut;
    }

    /**
     * Search the service description for an operation. If the operation is
     * found , checks if it has a child node named 'output'
     * 
     * @param serviceDesc
     *            dom Document representation of the service description
     * @param opName
     *            Name of the operation to search
     * @return true if the operation has been found and has the output child
     *         node or else false
     * @throws WSDLException
     * @throws TransformerException 
     */
    public static boolean isInOutOperation(Document serviceDesc, String opName,
            QName serviceName) throws WSDLException, TransformerException {
    	String docString = XMLHelper.createStringFromDOMDocument(serviceDesc);
        Document tmpDoc = XMLHelper.createDocumentFromString(docString);
        Definition serviceDef = createDefinitionFromDocument(tmpDoc);
        return isInOutOperation(serviceDef, opName, serviceName);
    }

    protected static Document createDocumentFromDefinition(Definition def)
        throws WSDLException {
        Document result = null;
        WSDLWriter writer = new WSDLWriterImpl();
        OutputStream outStream = new ByteArrayOutputStream();
        writer.writeWSDL(def, outStream);
        String documentToString = outStream.toString();
        result = XMLHelper.createDocumentFromString(documentToString);
        return result;
    }

    @SuppressWarnings("unchecked")
    protected static BindingOperation findOperationNamed(Definition serviceDef,
            String opName, QName serviceName) {
        BindingOperation result = null;
        boolean found = false;
        if (serviceDef != null) {
            Service service = serviceDef.getService(serviceName);
            if (service != null) {
                Map<String, Port> ports = service.getPorts();
                for (Port port : ports.values()) {
                    List<BindingOperation> ops = port.getBinding()
                            .getBindingOperations();
                    for (BindingOperation op : ops) {
                        if (op.getName().equals(opName)) {
                            result = op;
                            found = true;
                            break;
                        }
                    }
                    if (found) {
                        break;
                    }
                }
            }
        }
        return result;
    }

}
