/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: HelloworldImpl.java 1120 2006-09-21 15:36:50Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.helloworld;

import java.io.File;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NamingException;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.listener.MessageExchangeListener;
import org.objectweb.petals.component.common.logging.ComponentLogger;
import org.objectweb.petals.component.common.xmlmanipulation.StringHelper;
import org.objectweb.petals.component.common.xmlmanipulation.WSDLHelper;
import org.objectweb.petals.component.common.xmlmanipulation.XMLHelper;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Node;

/**
 * This component is a simple HelloworldImpl service engine
 * 
 * @author alouis, ddesjardins - eBMWebsourcing
 */
public class HelloworldImpl implements Helloworld, Component,
    ComponentLifeCycle {

    /**
     * Used to store in jndi a message
     */
    protected static String HELLOWORLD_CONTEXT = "helloworldContext";

    protected static String HELLOWORLD_JNDI_ENTRY = "helloworldEntry";

    private static final String ENDPOINT = "HelloworldEndpoint";

    private static final String SERVICE = "HelloworldService";

    private static final String TARGET_NS = "http://petals.objectweb.org/";

    private static final QName SERVICE_NAME = new QName(TARGET_NS, SERVICE);

    protected DeliveryChannel channel;

    private ComponentContext context;

    private ServiceEndpoint endpointReference;

    private MessageExchangeListener listener;

    private Logger logger;

    public ObjectName getExtensionMBeanName() {
        return null;
    }

    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    /**
     * If arg0.service = HelloworldService, return the wsdl description
     */
    public Document getServiceDescription(ServiceEndpoint arg0) {
        if (SERVICE_NAME.equals(arg0.getServiceName())) {
            File wsdlFile = new File(context.getInstallRoot() + File.separator
                + "META-INF" + File.separator + "HelloworldService.wsdl");
            return WSDLHelper.createDocumentFromWSDL(wsdlFile);
        }
        return null;
    }

    public ServiceUnitManager getServiceUnitManager() {
        return null;
    }

    public void init(ComponentContext ctx) throws JBIException {
        this.context = ctx;
        Logger log = ctx.getLogger("", "hello");
        logger = new ComponentLogger(log, log.getName(), log
            .getResourceBundleName(), context.getComponentName());
        channel = ctx.getDeliveryChannel();
        logger.log(Level.INFO, "hello.init");
    }

    public boolean isExchangeWithConsumerOkay(ServiceEndpoint arg0,
        MessageExchange arg1) {
        logger.log(Level.INFO, "hello.accept");
        return true;
    }

    public boolean isExchangeWithProviderOkay(ServiceEndpoint arg0,
        MessageExchange arg1) {
        logger.log(Level.INFO, "hello.refuse");
        return false;
    }

    /**
     * If requested service is HelloworldService, return the endpoint
     */
    public ServiceEndpoint resolveEndpointReference(DocumentFragment arg0) {

        ServiceEndpoint result = null;

        if (arg0 != null) {

            Node service = XMLHelper.findChild(arg0, "service", true);

            if (service != null) {

                String serviceRequested = XMLHelper.getAttributeValue(service,
                    "name");

                if (SERVICE.equals(serviceRequested)) {

                    result = endpointReference;
                }
            }
        }
        return result;
    }

    /**
     * 
     */
    public void shutDown() throws JBIException {

        logger.log(Level.INFO, "hello.shutdown");

        listener = null;
    }

    /**
     * 
     */
    public void start() throws JBIException {

        logger.log(Level.INFO, "hello.start");

        // Start the message processor
        try {

            HelloworldProcessor processor = new HelloworldProcessor(channel,
                logger, this);

            listener = new MessageExchangeListener(channel, processor,
                MessageExchangeListener.IgnoredStatus.DONE_AND_ERROR_IGNORED,
                0, context.getComponentName());

            listener.listen();

            endpointReference = this.context.activateEndpoint(SERVICE_NAME,
                ENDPOINT);

        } catch (MessagingException e) {
            throw new JBIException(e);
        }

        // Register in Jndi the Helloworld entry
        try {

            context.getNamingContext().createSubcontext(HELLOWORLD_CONTEXT);

        } catch (NameAlreadyBoundException e) {
            logger.log(Level.INFO,
                "the HelloWorld jndi context is already bound.");
        } catch (NamingException e) {
            throw new JBIException(e);
        }

    }

    public void stop() throws JBIException {

        logger.log(Level.INFO, "hello.stop");

        this.listener.terminate();
    }

    // ////////////////////////////////////////
    // interface implementation
    // ////////////////////////////////////////

    public void bindMessage(String message) throws Exception {

        if (StringHelper.isEmpty(message)) {

            logger.log(Level.WARNING, "hello.addNothing");

            // this will be automaticly added to a Fault if the message
            throw new Exception(
                "bindEntryInComponentInitialContext operation : you add nothing !!");
        }
        String jndiName = HelloworldImpl.HELLOWORLD_CONTEXT + "/"
            + HELLOWORLD_JNDI_ENTRY;

        context.getNamingContext().rebind(jndiName, message);
    }

    public void printMessage(String message) throws Exception {

        if (StringHelper.isEmpty(message)) {
            logger.log(Level.WARNING, "hello.printNothing");

            // this will be automaticly added to a Fault if the message
            throw new Exception(
                "printMessage operation : you told me nothing !!");
        }
        logger.log(Level.INFO, "hello.received", new String[] {message});
    }

    public String sayHello(String message) throws Exception {

        logger.log(Level.INFO, "hello.received", new String[] {message});

        if (StringHelper.isEmpty(message)) {

            logger.log(Level.WARNING, "hello.sayNothing");

            // this will be automaticly added to a Fault if the message
            throw new Exception("sayHello operation : you told me nothing !!");
        }
        ResourceBundle.getBundle("hello").getString("hello");

        String response = "<sayHelloResponse xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" "
            + "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"><sayHelloReturn xsi:type=\"xsd:string\">"
            + ResourceBundle.getBundle("hello").getString("hello")
            + " "
            + message + "' </sayHelloReturn></sayHelloResponse>";

        logger.log(Level.INFO, "hello.answer", new String[] {response});

        return response;
    }

}
