/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id:  SMTPServiceUnitManager.java 154 2006-03-27 15:30:10Z rbarraza $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.smtpbc.listeners;

import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Set;

import javax.activation.DataHandler;
import javax.mail.BodyPart;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Multipart;

import org.ebmwebsourcing.b2bgw.security.beans.CertificatePKCS12;
import org.objectweb.petals.binding.exception.SMTPBCException;

import com.isnetworks.smime.CertificateNotPresentException;
import com.isnetworks.smime.EnvelopedBodyPart;
import com.isnetworks.smime.SMIMEException;
import com.isnetworks.smime.SignedContent;

/**
 * This class allows to extract mail content, decrypt messages, and verify sign
 * messages
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Rocio BARRAZA - eBMWebsourcing
 * 
 */
public final class MessageHandler {

    private MessageHandler() {
        super();
    }

    /**
     * decrypte envelopped mime message,
     * 
     * @param envelopedBodyPart
     * @param cert
     *            receiver certificate
     * @param privateKey
     *            receiver private key
     * @return returns message content
     * @throws SMTPBCException
     */
    public static Object decryptContent(EnvelopedBodyPart envelopedBodyPart,
        X509Certificate cert, PrivateKey privateKey) throws SMTPBCException {
        Object decryptedContent = null;
        try {
            decryptedContent = envelopedBodyPart.getUnencryptedContent(cert,
                privateKey);
        } catch (CertificateNotPresentException e) {
            System.out.println("Not encrypted to the recipient certificate");
        } catch (NoSuchAlgorithmException e) {
            throw new SMTPBCException("Error decrypting message");
        } catch (MessagingException e) {
            throw new SMTPBCException("Error in crypted message");
        }
        return decryptedContent;
    }

    /**
     * "HandleMessage()" verify the type of message Mime and process it .
     * 
     * @throws SMTPBCException
     */
    public static String handleMessage(Message message,
        CertificatePKCS12 receiptCert, Set<DataHandler> attach)
        throws SMTPBCException {
        try {

            // String subject = message.getSubject();

            Object content = message.getContent();

            // Check if the message was signed
            if ((content instanceof SignedContent)
                || (content instanceof EnvelopedBodyPart)) {

                if (content instanceof SignedContent) {

                    SignedContent scontent = (SignedContent) content;

                    // cert is the sender certificate
                    content = MessageHandler.verifySignature(
                        (SignedContent) content,
                        takeSenderCertificate(scontent));
                }

                // Check if the message was encrypted (or encrypted, then
                // signed)
                if (content instanceof EnvelopedBodyPart) {
                    if (receiptCert == null) {
                        throw new Exception("Receipt certificate is null");
                    } else {
                        // cert is the receiver certificate, and PrivateKey is
                        // receiver private key
                        content = MessageHandler.decryptContent(
                            (EnvelopedBodyPart) content, receiptCert.getCert(),
                            receiptCert.getPrivateKey());
                    }
                }

                // Check if the message was signed, then encrypted
                if (content instanceof SignedContent) {
                    SignedContent scontent = (SignedContent) content;

                    // cert is the sender certificate
                    content = MessageHandler.verifySignature(scontent,
                        takeSenderCertificate(scontent));

                }
            }
            if (content instanceof Multipart) {
                Multipart multipart = (Multipart) content;
                content = MessageHandler.MultipartContent(multipart, attach);
            } else {
                content = message.getContent();
            }

            if (content != null) {
                return content.toString();
            } else {
                return null;
            }
        } catch (Exception ex) {
            throw new SMTPBCException(ex.getMessage());
        }
    }

    /**
     * 
     * @param messagePart
     *            it contents body of Mime message
     * @return
     * @throws SMTPBCException
     */
    public static Object MultipartContent(Multipart multipart,
        Set<DataHandler> attach) throws SMTPBCException {
        try {
            // -- Get the content of all messages multipart messages neither
            // crypted nor signed--
            int numberOfBody = multipart.getCount();
            BodyPart part = multipart.getBodyPart(0);
            DataHandler handler = null;
            Object content = "";

            if (part != null) {
                content = part.getContent();
            }
            for (int i = 0; i < numberOfBody; i++) {
                part = multipart.getBodyPart(i);
                handler = part.getDataHandler();
                if (handler.getName() != null) {
                    attach.add(handler);
                }
            }
            if (content != null) {
                return content.toString();
            } else {
                return "";
            }
        } catch (Exception ex) {
            throw new SMTPBCException(ex.getMessage());
        }
    }

    /**
     * It pulls out sender certificate that is attached to Mime message
     * 
     * @param scontent
     * @return X509Certificate, sender certificate
     * @throws SMIMEException
     */
    public static X509Certificate takeSenderCertificate(SignedContent scontent)
        throws SMIMEException {
        X509Certificate senderCert = null;
        Enumeration enumm = scontent.getCertificates();
        ArrayList<X509Certificate> listCert = new ArrayList<X509Certificate>();

        while (enumm.hasMoreElements()) {
            listCert.add((X509Certificate) enumm.nextElement());
        }

        if (listCert.size() > 0) {
            int i = 0;
            for (X509Certificate certificate : listCert) {
                if (certificate.getBasicConstraints() == -1) {
                    senderCert = listCert.get(i);
                    break;
                }
                i++;
            }
        }

        return senderCert;
    }

    /**
     * Verify signed message. The sender certificate is attached to the mail
     * 
     * @param signedContent
     *            Mime signed part
     * @param cert
     *            sender certificate
     * @return returns message content
     * @throws SMTPBCException
     *             sender certificate don't present, Error verifying signed
     *             message
     */
    public static Object verifySignature(SignedContent signedContent,
        X509Certificate cert) throws SMTPBCException {
        try {
            boolean verified = signedContent.verify(cert);
            if (!verified) {
                System.out.println("Signature don't verify \n");
            }
            return signedContent.getSignedContent();

        } catch (CertificateNotPresentException e) {
            throw new SMTPBCException("sender certificate don't present ", e
                .getMessage());
        } catch (Exception ex) {
            throw new SMTPBCException("Error verifying signed message", ex
                .getMessage());
        }
    }
}
