/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: AutoLoaderImpl.java,v 1.2 2006/04/07 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.management.autoload;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Timer;

import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.apache.commons.io.FileUtils;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.petals.jbi.management.service.DeploymentService;
import org.objectweb.petals.jbi.management.service.InstallationService;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.management.service.PackageHandler;
import org.objectweb.petals.jbi.management.service.util.XMLResult;
import org.objectweb.petals.jbi.management.systemstate.ComponentState;
import org.objectweb.petals.jbi.management.systemstate.ServiceAssemblyState;
import org.objectweb.petals.jbi.management.systemstate.SharedLibraryState;
import org.objectweb.petals.jbi.management.systemstate.SystemState;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_DEPLOYMENT_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_INSTALLATION_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_LIFECYCLEMANAGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_SYSTEMSTATE_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.PLATFORM_REPOSITORY_NAME;

/**
 * This class is used for automaticly loading components which are placed in the
 * "autoLoad" directory. It periodicly scan the directory and check if new
 * components are presents. The InstallationService is called for each of those
 * components. The whole process of installation is done (install + start) After
 * being threat, the archives of the components are removed. A file
 * "fileName.success" will be created if success A file "fileName.failed" will
 * be created otherwise, containing the related exception
 * 
 * @author alouis
 * 
 */
public class AutoLoaderImpl implements AutoLoader, BindingController,
    LifeCycleController {

    protected static final String[] CLIENT_INTERFACES = new String[] {
        JBI_DEPLOYMENT_NAME, JBI_INSTALLATION_NAME, JBI_LIFECYCLEMANAGER_NAME,
        JBI_SYSTEMSTATE_NAME};

    private static String install = "install";

    protected static File installDirectory;

    private static String installed = "installed";

    protected static File installedDirectory;

    private static String uninstalled = "uninstalled";

    protected static File uninstalledDirectory;

    private static String work = "work";

    protected static File workDirectory;

    protected DeploymentService deploymentService;

    protected InstallationService installationService;

    protected Timer installDirectoryScanTimer;

    protected Timer installedDirectoryScanTimer;

    protected LoggingUtil log;

    protected Logger logger;

    protected LifeCycleManagerService managerService;

    /**
     * Package Handler
     */
    protected PackageHandler packageHandler;

    /**
     * JBI Container :: Recovery Service
     */
    protected SystemState recoverySrv;

    /**
     * Platform Component :: Repository Service
     */
    protected RepositoryService repositorySrv;

    /**
     * Component LifeCycle state.
     */
    private String fcState;

    private long scanPeriod = 1000L;

    public static File getInstallDirectory() {
        return installDirectory;
    }

    public static File getInstalledDirectory() {
        return installedDirectory;
    }

    public static File getUninstalledDirectory() {
        return uninstalledDirectory;
    }

    public static File getWorkDirectory() {
        return workDirectory;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
     *      java.lang.Object)
     */
    public void bindFc(String clientItfName, Object serverItf)
        throws NoSuchInterfaceException, IllegalBindingException,
        IllegalLifeCycleException {

        if (LOGGER_NAME.equals(clientItfName)) {
            logger = (Logger) serverItf;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else if (JBI_INSTALLATION_NAME.equals(clientItfName)) {
            installationService = (InstallationService) serverItf;
        } else if (JBI_DEPLOYMENT_NAME.equals(clientItfName)) {
            deploymentService = (DeploymentService) serverItf;
        } else if (JBI_LIFECYCLEMANAGER_NAME.equals(clientItfName)) {
            managerService = (LifeCycleManagerService) serverItf;
        } else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
            repositorySrv = (RepositoryService) serverItf;
        } else if (JBI_SYSTEMSTATE_NAME.equals(clientItfName)) {
            recoverySrv = (SystemState) serverItf;

        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
     */
    public String getFcState() {
        return STARTED.equals(fcState) ? STARTED : STOPPED;
    }

    /**
     * init the autoloader : init the directories
     */
    public void init() throws IOException {
        log = new LoggingUtil(logger);
        log.call();
        installDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + install);
        installedDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + installed);
        uninstalledDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + uninstalled);
        workDirectory = new File(SystemUtil.getPetalsInstallDirectory()
            .getAbsolutePath()
            + File.separator + work);

        checkValidiyOfdirectory(installDirectory, false);
        checkValidiyOfdirectory(installedDirectory, false);
        checkValidiyOfdirectory(uninstalledDirectory, false);
        checkValidiyOfdirectory(workDirectory, true);

    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#listFc()
     */
    public String[] listFc() {
        return CLIENT_INTERFACES;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
     */
    public Object lookupFc(String clientItfName)
        throws NoSuchInterfaceException {
        if (LOGGER_NAME.equals(clientItfName)) {
            return logger;
        } else if (JBI_INSTALLATION_NAME.equals(clientItfName)) {
            return installationService;
        } else if (JBI_DEPLOYMENT_NAME.equals(clientItfName)) {
            return deploymentService;
        } else if (JBI_LIFECYCLEMANAGER_NAME.equals(clientItfName)) {
            return managerService;
        } else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
            return repositorySrv;
        } else if (JBI_SYSTEMSTATE_NAME.equals(clientItfName)) {
            return recoverySrv;

        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * start the autoloader : launch the scan
     */
    public void start() {
        log.call();

        installDirectoryScanTimer = new Timer();
        installedDirectoryScanTimer = new Timer();
        InstallDirectoryScanner installScan = new InstallDirectoryScanner(this,
            installDirectory);
        InstalledDirectoryScanner installedScan = new InstalledDirectoryScanner(
            this, installedDirectory);
        /*
         * Initialize package handler
         */
        packageHandler = new PackageHandler(logger, repositorySrv);

        installDirectoryScanTimer.schedule(installScan, scanPeriod, scanPeriod);
        installedDirectoryScanTimer.schedule(installedScan, scanPeriod,
            scanPeriod);
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    public void startFc() throws IllegalLifeCycleException {

        try {
            init();
        } catch (IOException e) {
            log.error(e);
            throw new IllegalLifeCycleException(e.getMessage());
        }
        start();
        fcState = STARTED;
    }

    /**
     * Stop the autoload of the components. TODO : synchronization, does the
     * autoload must wait before the end of an installation before closing?
     */
    public void stop() {
        log.call();

        if (installDirectoryScanTimer != null) {
            installDirectoryScanTimer.cancel();
        }
        if (installedDirectoryScanTimer != null) {
            installedDirectoryScanTimer.cancel();
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
     */
    public void stopFc() throws IllegalLifeCycleException {
        stop();
        fcState = STOPPED;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
     */
    public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
        IllegalBindingException, IllegalLifeCycleException {

        if (LOGGER_NAME.equals(clientItfName)) {
            logger = null;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else if (JBI_INSTALLATION_NAME.equals(clientItfName)) {
            installationService = null;
        } else if (JBI_DEPLOYMENT_NAME.equals(clientItfName)) {
            deploymentService = null;
        } else if (JBI_LIFECYCLEMANAGER_NAME.equals(clientItfName)) {
            managerService = null;
        } else if (PLATFORM_REPOSITORY_NAME.equals(clientItfName)) {
            repositorySrv = null;
        } else if (JBI_SYSTEMSTATE_NAME.equals(clientItfName)) {
            recoverySrv = null;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * friendly method for the directoryscanner
     * 
     * @param newInstallers
     */
    protected void install(File[] newInstallers) {
        log.call();

        if (newInstallers != null) {
            for (int i = 0; i < newInstallers.length; i++) {
                File file = newInstallers[i];
                try {
                    /*
                     * Copy archive file to work dir
                     */
                    File workArchive = new File(getWorkDirectory(), file
                        .getName());
                    FileUtils.copyFile(file, workArchive);
                    FileUtils.forceDelete(file);

                    /*
                     * Load JBI descriptor here
                     */

                    JBIDescriptor descriptor = packageHandler
                        .loadDescriptor(workArchive.toURI());

                    /*
                     * Perform the right task : install, share or deploy
                     */
                    if (descriptor.getComponent() != null) {
                        performInstall(workArchive);
                    } else if (descriptor.getSharedLibrary() != null) {
                        performShare(workArchive);
                    } else if (descriptor.getServiceAssembly() != null) {
                        performDeploy(workArchive);
                    } else {
                        throw new Exception("auto recover of entity failed : "
                            + workArchive.getCanonicalPath());
                    }
                    FileUtils.forceDelete(workArchive);

                } catch (Throwable e) {
                    log.error("Error during install", e);
                }
            }
        }
    }

    /**
     * friendly method for the directoryscanner
     * 
     * @param newInstallers
     */
    protected void uninstall(List<File> filesToUninstall) {
        log.call();
        /*
         * Retrieve all installed entities
         */
        List<ComponentState> compStates = recoverySrv
            .retrieveAllComponentStates();
        List<SharedLibraryState> slStates = recoverySrv
            .retrieveAllSharedLibraryStates();
        List<ServiceAssemblyState> saStates = recoverySrv
            .retrieveAllServiceAssemblyStates();

        /*
         * Check if one of the files to uninstall match one of the archiveUrl of
         * the installed entities. If yes, this entity must be uninstalled.
         */
        List<String> compToUninstall = new ArrayList<String>();
        List<String> slToUninstall = new ArrayList<String>();
        List<String> saToUninstall = new ArrayList<String>();
        for (ComponentState state : compStates) {
            URI archiveURI = packageHandler.processAndGetPackageURI(state
                .getArchiveURL(), false);
            File archiveFile = new File(archiveURI);
            for (File file : filesToUninstall) {
                if (file.equals(archiveFile)) {
                    compToUninstall.add(state.getName());
                }
            }
        }
        for (SharedLibraryState state : slStates) {
            URI archiveURI = packageHandler.processAndGetPackageURI(state
                .getArchiveURL(), false);
            File archiveFile = new File(archiveURI);
            for (File file : filesToUninstall) {
                if (file.equals(archiveFile)) {
                    slToUninstall.add(state.getName());
                }
            }
        }
        for (ServiceAssemblyState state : saStates) {
            URI archiveURI = packageHandler.processAndGetPackageURI(state
                .getArchiveURL(), false);
            File archiveFile = new File(archiveURI);
            for (File file : filesToUninstall) {
                if (file.equals(archiveFile)) {
                    saToUninstall.add(state.getName());
                }
            }
        }

        /*
         * Perform entity uninstallation
         */
        for (String name : saToUninstall) {
            performSAUninstall(name);
        }
        for (String name : compToUninstall) {
            performCompUninstall(name);
        }
        for (String name : slToUninstall) {
            performSLUninstall(name);
        }

    }

    // ----------------------------------------------------------------------
    // Fractal BindingController implementation
    // ----------------------------------------------------------------------

    /**
     * Check if the specified File is a directory and is created. If it does not
     * exist, create it.
     * 
     * @param dir
     * @param clean
     *            delete directory content or not
     */
    private void checkValidiyOfdirectory(File dir, boolean clean)
        throws IOException {
        if (dir != null) {
            // parameter is not null
            if (!dir.exists()) {
                // directory does not exist, create it
                if (!dir.mkdirs()) {
                    // creation failed
                    throw new IOException(
                        "can not create the following directory : "
                            + dir.getCanonicalPath());
                }
            } else {
                // file already exist, check it is a directory
                if (!dir.isDirectory()) {
                    // file is not a directory
                    throw new IOException(dir.getCanonicalPath()
                        + " already exists and is not a directory.");
                } else {
                    // clean the directory
                    if (clean) {
                        FileUtils.deleteDirectory(dir);
                        dir.mkdirs();
                    }
                }
            }
        } else {
            throw new FileNotFoundException("the specified directory is null");
        }
    }

    /**
     * Uninstall the specified component.
     * 
     * @param name
     */
    protected void performCompUninstall(String name) {
        installationService.unloadInstaller(name, true);
    }

    /**
     * Deploy the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     * 
     * @param f
     */
    private void performDeploy(File f) {
        log.call();

        if (f != null && f.exists()) {

            try {

                // TODO Find a better way of managing URI
//                String absPath = f.getAbsolutePath().replaceAll(" ", "%20");
//                String archiveURI;
//                if (absPath.startsWith("/")) {
//                    archiveURI = "file://" + absPath;
//                } else {
//                    archiveURI = "file:///" + absPath;
//                }

                String xml = deploymentService.deploy(f.toURI().toString());

                if (xml.indexOf(XMLResult.TaskResult.FAILED.toString()) > -1) {
                    log.warning("Problem while deploying a service assembly\n"
                        + xml);
                } else {
                    String saName = "";
                    saName = xml.substring(xml.indexOf("<loc-param>") + 11, xml
                        .indexOf("</loc-param>"));
                    deploymentService.start(saName);
                    
                }
            } catch (Throwable e) {
                try {
                    File failed = new File(workDirectory, f.getName()
                        + ".failed");

                    failed.createNewFile();
                    PrintWriter w = null;
                    w = new PrintWriter(failed);
                    e.printStackTrace(w);
                    w.flush();
                    w.close();
                    log.error(e);
                } catch (IOException e1) {
                    log.error(e1);
                }
            }
        }
    }

    /**
     * Install the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     * 
     * @param f
     */
    private void performInstall(File f) {
        log.call();

        if (f != null && f.exists()) {

            try {

                MBeanServer jmxServer = managerService.getMBeanServer();

                // TODO Find a better way of managing URI
                // String absPath = f.getAbsolutePath().replaceAll(" ", "%20");
                // String archiveURI;
                // if (absPath.startsWith("/")) {
                // archiveURI = "file://" + absPath;
                // } else {
                // archiveURI = "file:///" + absPath;
                // }

                ObjectName newInstallerName = installationService
                    .loadNewInstaller(f.toURI().toString());
                if (!newInstallerName.getDomain().equals("error")) {
                    Object[] nullObjects = new Object[0];
                    String[] nullStrings = new String[0];

                    Object result = jmxServer.invoke(newInstallerName,
                        "install", nullObjects, nullStrings);

                    if (result instanceof ObjectName) {
                        ObjectName componentName = (ObjectName) result;

                        jmxServer.invoke(componentName, "start", nullObjects,
                            nullStrings);

                    } else {
                        throw new Exception("error during installation");
                    }
                } else {
                    throw new Exception("Component can't be autoinstalled");
                }
            } catch (Throwable e) {
                try {

                    File failed = new File(workDirectory, f.getName()
                        + ".failed");

                    failed.createNewFile();
                    PrintWriter w = null;
                    w = new PrintWriter(failed);
                    e.printStackTrace(w);
                    w.flush();
                    w.close();
                    log.error(e);
                } catch (IOException e1) {
                    log.error(e1);
                }
            }
        }
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    /**
     * Undeploy the specified service assembly. isToBeDeleted is setted to true !
     * 
     * @param name
     */
    protected void performSAUninstall(String name) {
        try {
            deploymentService.undeploy(name);
        } catch (Exception e) {
            log.error(e);
        }
    }

    /**
     * Install the specified shared library in the JBI container Any exception
     * will cause the abortion of the process. A file "fileName.success" will be
     * created if success A file "fileName.failed" will be created otherwise,
     * containing the related exception
     * 
     * @param f
     */
    private void performShare(File f) {
        log.call();

        if (f != null && f.exists()) {

            try {

                // TODO Find a better way of managing URI
//                String absPath = f.getAbsolutePath().replaceAll(" ", "%20");
//                String archiveURI;
//                if (absPath.startsWith("/")) {
//                    archiveURI = "file://" + absPath;
//                } else {
//                    archiveURI = "file:///" + absPath;
//                }

                installationService.installSharedLibrary(f.toURI().toString());


            } catch (Throwable e) {
                try {

                    File failed = new File(workDirectory, f.getName()
                        + ".failed");

                    failed.createNewFile();
                    PrintWriter w = null;
                    w = new PrintWriter(failed);
                    e.printStackTrace(w);
                    w.flush();
                    w.close();
                    log.error(e);
                } catch (IOException e1) {
                    log.error(e1);
                }
            }
        }
    }

    /**
     * Uninstall the specified shared lib.
     * 
     * @param name
     */
    protected void performSLUninstall(String name) {
        installationService.uninstallSharedLibrary(name);
    }

}
