/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: Manager.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.management.service;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.jmx.agent.AdminAttributes;
import org.objectweb.petals.jbi.management.lifecycle.ComponentLifeCycle;
import org.objectweb.petals.jbi.management.lifecycle.Installer;
import org.objectweb.petals.jbi.management.lifecycle.LifeCycleAbstract;
import org.objectweb.petals.jbi.management.lifecycle.ServiceAssemblyLifeCycle;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

import static org.objectweb.petals.util.InterfaceNamesHelper.JMXAGENT_ADMINATT_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.MONOLOG_FACTORY_NAME;

/**
 * <code>Manager</code> is the main class for JMX management. It is the entry
 * point for register and unregister elements in the JMX environment (elements
 * are JBI binding/engine components and systemServices components). Related
 * MBeans are created when registering and unregistering elements.
 * 
 * <code>Manager</code> owns the JMX management services offered by the JBI
 * container to administrative clients. Management services are : - AdminService
 * (referenced in the JMX server as "admin"); - InstallationService (referenced
 * in the JMX server as "installation"); - DeploymentService (referenced in the
 * JMX server as "deployment");
 * 
 * <code>Manager</code> is the central point for accessing the mBeans
 * associated to the JBI container elements.
 * 
 * <code>ManagementException</code> exceptions are thrown when problems for
 * accessing the JMX server occur.
 * 
 * TODO refactor to get System Services from Fractal, and BC and SE from JMX
 * TODO set all properties in a single file, and add a FractalAttribute
 * component to retrieve it (for joram, container name, etc)
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */

public class LifeCycleManagerImpl implements LifeCycleManagerService,
    BindingController, LifeCycleController {

    private static final String[] CLIENT_INTERFACES = new String[] {JMXAGENT_ADMINATT_NAME};

    private static final String DEPLOYMENT_NAME = "deployment";

    private static final String INSTALLATION_NAME = "installation";

    protected ObjectName addressName;

    protected ObjectName adminName;

    protected Map<ObjectName, ComponentLifeCycle> bindingCompoLifeCycles;

    protected String containerName;

    protected ObjectName deploymentName;

    protected ObjectName directoryName;

    protected Map<ObjectName, ComponentLifeCycle> engineCompoLifeCycles;

    protected ObjectName installationName;

    protected Map<ObjectName, Installer> installers;

    protected ObjectName itineraryName;

    /**
     * Monolog logger
     */
    protected Logger logger;

    protected MBeanNamesImpl mBeanNames;

    protected AdminAttributes adminAttributes;

    protected Map<String, ServiceAssemblyLifeCycle> serviceAssemblyLifeCycles;

    protected Map<ObjectName, LifeCycleAbstract> systemServices;

    protected ObjectName transporterName;

    /**
     * Component lifecycle state.
     */
    private String fcState;

    /**
     * Logger wrapper
     */
    private LoggingUtil log;

    /**
     * Default constructor. Creates a <code>Manager</code> instance. JMX
     * server is created.
     * 
     */
    public LifeCycleManagerImpl() {

        mBeanNames = new MBeanNamesImpl("org.objectweb.petals");

        bindingCompoLifeCycles = Collections
            .synchronizedMap(new HashMap<ObjectName, ComponentLifeCycle>());

        engineCompoLifeCycles = Collections
            .synchronizedMap(new HashMap<ObjectName, ComponentLifeCycle>());

        systemServices = Collections
            .synchronizedMap(new HashMap<ObjectName, LifeCycleAbstract>());

        serviceAssemblyLifeCycles = Collections
            .synchronizedMap(new HashMap<String, ServiceAssemblyLifeCycle>());

        installers = Collections
            .synchronizedMap(new HashMap<ObjectName, Installer>());

        deploymentName = mBeanNames
            .createSystemComponentMBeanName(DEPLOYMENT_NAME);

        installationName = mBeanNames
            .createSystemComponentMBeanName(INSTALLATION_NAME);
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
     *      java.lang.Object)
     */
    public void bindFc(String clientItfName, Object serverItf)
        throws NoSuchInterfaceException, IllegalBindingException,
        IllegalLifeCycleException {
        if (LOGGER_NAME.equals(clientItfName)) {
            logger = (Logger) serverItf;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else if (JMXAGENT_ADMINATT_NAME.equals(clientItfName)) {
            adminAttributes = (AdminAttributes) serverItf;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    // /////////////////////////
    // /// retrieve platform component MBeanNames
    // /////////////////////////
    /**
     * Retrieve the AdminServieMBean ObjectName.
     */
    public ObjectName getAdminServiceMBeanName() throws ManagementException {
        ObjectName result = null;
        String pattern = "FC/Petals/jbi/admin-impl@*:itf=service";
        Set objNames;
        try {
            ObjectName objName = new ObjectName(pattern);
            objNames = adminAttributes.getRawMBeanServer().queryNames(objName,
                null);
        } catch (MalformedObjectNameException e) {
            String msg = "Error during AdminService ObjName retrieval";
            log.error(msg);
            throw new ManagementException(msg, e);
        } catch (NullPointerException e) {
            String msg = "Error during AdminService ObjName retrieval";
            log.error(msg);
            throw new ManagementException(msg, e);
        }
        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        } else {
            throw new ManagementException("AdminService Mbean can't be "
                + "correctly retrieved with this pattern :" + pattern);
        }
        return result;
    }

    /**
     * @see org.objectweb.petals.jbi.management.service.LifeCycleManagerService#getComponentByName(java.lang.String)
     */
    public ComponentLifeCycle getComponentByName(String name) {
        log.call();

        ComponentLifeCycle on = null;

        // search in bindings

        on = getBindingComponentByName(name);

        if (on == null) {
            // search in engines

            on = getEngineComponentByName(name);
        }
        return on;
    }

    /**
     * Retrieve the DeploymentService ObjectName.
     * This is used by a ServiceAssembly LifeCycle.
     * As there is no MBean for the SA LifeCycle,
     * the DeploymentServiceMbean is used to represent it.
     * All SA notifications are reported to the DeploymentLifeCycleMB
     * 
     *  @return the jmx name of the deploymentServiceMBean
     */
    public ObjectName getDeploymentServiceMBeanName()
        throws ManagementException {
        ObjectName result = null;
        String pattern = "Petals:name=Deployment,type=service";
        Set objNames;
        try {
            ObjectName objName = new ObjectName(pattern);
            objNames = adminAttributes.getRawMBeanServer().queryNames(objName,
                null);
        } catch (MalformedObjectNameException e) {
            String msg = "Error during DeploymentService ObjName retrieval";
            log.error(msg);
            throw new ManagementException(msg, e);
        } catch (NullPointerException e) {
            String msg = "Error during DeploymentService ObjName retrieval";
            log.error(msg);
            throw new ManagementException(msg, e);
        }
        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        } else {
            throw new ManagementException("DeploymentService Mbean can't be "
                + "correctly retrieved with this pattern :" + pattern);
        }
        return result;
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
     */
    public String getFcState() {
        return STARTED.equals(fcState) ? STARTED : STOPPED;
    }

    /**
     * Retrieve the InstallationService ObjectName.
     */
    public ObjectName getInstallationServiceMBeanName()
        throws ManagementException {
        ObjectName result = null;
        String pattern = "FC/Petals/jbi/installation-impl@*:itf=service";
        Set objNames;
        try {
            ObjectName objName = new ObjectName(pattern);
            objNames = adminAttributes.getRawMBeanServer().queryNames(objName,
                null);
        } catch (MalformedObjectNameException e) {
            String msg = "Error during InstallationService ObjName retrieval";
            log.error(msg);
            throw new ManagementException(msg, e);
        } catch (NullPointerException e) {
            String msg = "Error during InstallationService ObjName retrieval";
            log.error(msg);
            throw new ManagementException(msg, e);
        }
        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        } else {
            throw new ManagementException("InstallationService Mbean can't be "
                + "correctly retrieved with this pattern :" + pattern);
        }
        return result;
    }

    // ////////////////////////////
    // manage components : bindings and engines
    // ////////////////////////////

    /**
     * Return the component installer associated with this componant name.
     * 
     * @param name
     *            the component name
     * @return The related installer, or null
     */
    @SuppressWarnings("unchecked")
    public ObjectName getInstallerByName(String name) {
        log.call();

        ObjectName result = null;

        Map<ObjectName, Installer> installers = getInstallers();

        synchronized (installers) {
            for (Iterator iter = installers.entrySet().iterator(); iter
                .hasNext()
                && result == null;) {
                Entry<ObjectName, Installer> entry = (Entry<ObjectName, Installer>) iter
                    .next();
                ObjectName objName = entry.getKey();
                Installer installer = entry.getValue();

                if (installer.getComponentName().equals(name)) {
                    result = objName;
                }
            }
        }
        return result;
    }

    /**
     * Return the <code>MBeanNamesImpl</code> element which allows the
     * creation of mBeans names.
     * 
     * @return
     */
    public MBeanNamesImpl getMBeanNames() {
        return mBeanNames;
    }

    /**
     * Return the JMX server used by the JBI container.
     * 
     * @return
     */
    public MBeanServer getMBeanServer() {
        return adminAttributes.getRawMBeanServer();
    }

    /**
     * Return the service assembly associated with this name.
     * 
     * @param saName
     * @return The related ServiceAssemblyLifeCycle, or null
     */
    public ServiceAssemblyLifeCycle getServiceAssemblyByName(String saName) {
        log.call();

        ServiceAssemblyLifeCycle result = null;

        Map<String, ServiceAssemblyLifeCycle> sas = this
            .getServiceAssemblyLifeCycles();

        synchronized (sas) {
            result = sas.get(saName);
        }
        return result;
    }

    /**
     * Return the synchronized map of binding components.
     * 
     * @return
     */
    public Map<ObjectName, ComponentLifeCycle> getBindingCompoLifeCycles() {
        return bindingCompoLifeCycles;
    }

    /**
     * Return the synchronized map of engine components.
     * 
     * @return
     */
    public Map<ObjectName, ComponentLifeCycle> getEngineCompoLifeCycles() {
        return engineCompoLifeCycles;
    }

    /**
     * Return the synchronized map of installers.
     * 
     * @return
     */
    public Map<ObjectName, Installer> getInstallers() {
        return installers;
    }

    /**
     * Return the synchronized map of deployed service asse mbly.
     * 
     * @return
     */
    public Map<String, ServiceAssemblyLifeCycle> getServiceAssemblyLifeCycles() {
        return serviceAssemblyLifeCycles;
    }

    /**
     * Return the synchronized map of system services components.
     * 
     * @return
     */
    public Map<ObjectName, LifeCycleAbstract> getSystemServices() {
        return systemServices;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#listFc()
     */
    public String[] listFc() {
        return CLIENT_INTERFACES;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
     */
    public Object lookupFc(String clientItfName)
        throws NoSuchInterfaceException {

        if (JMXAGENT_ADMINATT_NAME.equals(clientItfName)) {
            return adminAttributes;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * Register a mbean in the JMX server representing the specified binding
     * component.
     * 
     * @param componentElement
     * @throws ManagementException
     */
    public ObjectName registerBindingComponent(ComponentLifeCycle lifeCycle)
        throws ManagementException {

        ObjectName name = lifeCycle.getMBeanName();

        Map<ObjectName, ComponentLifeCycle> bindings = getBindingCompoLifeCycles();

        synchronized (bindings) {
            try {
                adminAttributes.getRawMBeanServer().registerMBean(lifeCycle,
                    name);

                bindings.put(name, lifeCycle);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }
        return name;
    }

    /**
     * Register a mbean in the JMX server representing the specified engine
     * component
     * 
     * @param componentElement
     * @throws ManagementException
     */
    public ObjectName registerEngineComponent(ComponentLifeCycle lifeCycle)
        throws ManagementException {
        ObjectName name = lifeCycle.getMBeanName();

        Map<ObjectName, ComponentLifeCycle> engines = getEngineCompoLifeCycles();

        synchronized (engines) {
            try {
                adminAttributes.getRawMBeanServer().registerMBean(lifeCycle,
                    name);

                engines.put(name, lifeCycle);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }
        return name;
    }

    // ////////////////////////////
    // manage installers
    // ////////////////////////////
    /**
     * Register an installer mbean in the JMX server representing the specified
     * component installer.
     * 
     * @param installerMBean
     * @throws ManagementException
     */
    public ObjectName registerInstaller(Installer installerMBean)
        throws ManagementException {

        ObjectName installerMBeanName = installerMBean.getMBeanName();

        Map<ObjectName, Installer> installers = getInstallers();

        synchronized (installers) {
            try {
                adminAttributes.getRawMBeanServer().registerMBean(
                    installerMBean, installerMBeanName);

                installers.put(installerMBeanName, installerMBean);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }
        return installerMBeanName;
    }

    // ////////////////////////////
    // manage service assemblies
    // ////////////////////////////
    /**
     * Register a ServiceAssemblyLifeCycle into the a synchronized Map<String,ServiceAssemblyLifeCycle>
     * 
     * @param saName
     * @param serviceAssemblyLifeCycle
     */
    public void registerServiceAssembly(String saName,
        ServiceAssemblyLifeCycle serviceAssemblyLifeCycle) {
        Map<String, ServiceAssemblyLifeCycle> sas = getServiceAssemblyLifeCycles();
        synchronized (sas) {
            sas.put(saName, serviceAssemblyLifeCycle);
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    public void startFc() throws IllegalLifeCycleException {

        log = new LoggingUtil(logger);

        fcState = STARTED;
    }

    // ----------------------------------------------------------------------
    // Fractal BindingController implementation
    // ----------------------------------------------------------------------

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
     */
    public void stopFc() throws IllegalLifeCycleException {

        fcState = STOPPED;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
     */
    public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
        IllegalBindingException, IllegalLifeCycleException {

        if (LOGGER_NAME.equals(clientItfName)) {
            logger = null;
        } else if (MONOLOG_FACTORY_NAME.equals(clientItfName)) {
            // nothing
        } else if (JMXAGENT_ADMINATT_NAME.equals(clientItfName)) {
            adminAttributes = null;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * Unregister the binding component of the JMX server
     * 
     * @throws ManagementException
     */
    public void unregisterBindingComponent(ObjectName name)
        throws ManagementException {
        Map bindings = getBindingCompoLifeCycles();

        synchronized (bindings) {
            try {
                adminAttributes.getRawMBeanServer().unregisterMBean(name);

                bindings.remove(name);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }
    }

    /**
     * Unregister the engine component of the JMX server
     * 
     * @throws ManagementException
     */
    public void unregisterEngineComponent(ObjectName name)
        throws ManagementException {
        Map engines = getEngineCompoLifeCycles();

        synchronized (engines) {
            try {
                adminAttributes.getRawMBeanServer().unregisterMBean(name);

                engines.remove(name);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    /**
     * Unregister the component installer of the JMX server
     * 
     * @throws ManagementException
     */
    public void unregisterInstaller(ObjectName name) throws ManagementException {
        Map<ObjectName, Installer> installers = getInstallers();

        synchronized (installers) {
            try {
                adminAttributes.getRawMBeanServer().unregisterMBean(name);

                installers.remove(name);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }
    }

    /**
     * Unregister the ServiceAssembly life cycle from the sync map
     * 
     * @throws ManagementException
     */
    public void unregisterServiceAssembly(String saName)
        throws ManagementException {
        Map sas = getServiceAssemblyLifeCycles();

        synchronized (sas) {
            try {
                sas.remove(saName);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }
    }

    /**
     * Return the binding component associated with this name.
     * 
     * @param name
     * @return The related objectName, or null
     */
    private ComponentLifeCycle getBindingComponentByName(String name) {
        log.call();

        ComponentLifeCycle result = null;

        Map bindings = this.getBindingCompoLifeCycles();

        synchronized (bindings) {
            for (Iterator iter = bindings.values().iterator(); iter.hasNext()
                && result == null;) {
                ComponentLifeCycle lifeCycle = (ComponentLifeCycle) iter.next();

                if (lifeCycle.getName().equals(name)) {
                    result = lifeCycle;
                }
            }
        }
        return result;
    }

    /**
     * Return the engine component associated with this name.
     * 
     * @param name
     * @return The related objectName, or null
     */
    private ComponentLifeCycle getEngineComponentByName(String name) {
        log.call();

        ComponentLifeCycle result = null;

        Map engines = this.getEngineCompoLifeCycles();

        synchronized (engines) {
            for (Iterator iter = engines.values().iterator(); iter.hasNext()
                && result == null;) {
                ComponentLifeCycle lifeCycle = (ComponentLifeCycle) iter.next();

                if (lifeCycle.getName().equals(name)) {
                    result = lifeCycle;
                }
            }
        }
        return result;
    }
}
