/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SADeploymentManager.java 154 31 juil. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.service;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.ZipFile;

import javax.jbi.JBIException;
import javax.jbi.management.LifeCycleMBean;
import javax.management.ObjectName;
import javax.xml.namespace.QName;

import org.apache.commons.io.FileUtils;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.management.autoload.AutoLoaderImpl;
import org.objectweb.petals.jbi.management.lifecycle.LifeCycleAbstract;
import org.objectweb.petals.jbi.management.lifecycle.ServiceAssemblyLifeCycle;
import org.objectweb.petals.jbi.management.service.util.XMLResult;
import org.objectweb.petals.jbi.management.service.util.XMLResult.CauseFramework;
import org.objectweb.petals.jbi.management.service.util.XMLResult.MessageType;
import org.objectweb.petals.jbi.management.service.util.XMLResult.TaskResult;
import org.objectweb.petals.jbi.management.systemstate.SystemState;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.repository.RepositoryImpl.EntityType;
import org.objectweb.petals.tools.jbicommon.descriptor.Connection;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceAssembly;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceUnit;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.StringHelper;
import org.objectweb.petals.util.ZipUtil;

/**
 * SADeploymentManager is used to handle the whole process of a specific service
 * assembly deployment.
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class SADeploymentManager {

    /**
     * The service assembly production element
     */
    protected ServiceAssembly assembly;

    /**
     * The platform deployment service
     */
    protected DeploymentServiceImpl deploymentService;

    /**
     * The platform endpoint service
     */
    protected EndpointService endpointService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * JBI Container :: JMX Agent Admin
     */
    protected LifeCycleManagerService managerService;

    /**
     * Platform util : package handler
     */
    protected PackageHandler packageHandler;

    /**
     * JBI Container :: Recovery Service
     */
    protected SystemState recoverySrv;

    /**
     * Platform Component :: Repository Service
     */
    protected RepositoryService repositorySrv;

    /**
     * The service assembly archive URI
     */
    protected URI saArchiveURI;

    /**
     * Installation folders of all SUs included into the managed SA
     */
    protected Map<String, URI> suInstallRoots = new HashMap<String, URI>();

    /**
     * The xml message manager
     */
    protected XMLResult xmlResult;

    public SADeploymentManager() {
        super();
    }

    /**
     * 
     * @param assembly
     *            the {@link ServiceAssembly} production element
     * @param saArchiveURI
     *            the service assembly archive {@link URI}
     * @param packageHandler
     *            the platform {@link PackageHandler} util
     * @param endpointService
     *            the platform {@link EndpointService}
     * @param deploymentService
     *            the platform {@link DeploymentServiceImpl}
     * @param repositorySrv
     *            the platform {@link RepositoryService}
     * @param recoverySrv
     *            the platform {@link SystemState} manager
     * @param log
     *            the platform {@link LoggingUtil}
     * @param managerService
     *            the platform {@link LifeCycleManagerService}
     */
    public SADeploymentManager(ServiceAssembly assembly, URI saArchiveURI,
            PackageHandler packageHandler, EndpointService endpointService,
            DeploymentServiceImpl deploymentService,
            RepositoryService repositorySrv, SystemState recoverySrv,
            LoggingUtil log, LifeCycleManagerService managerService) {
        super();
        this.assembly = assembly;
        this.saArchiveURI = saArchiveURI;
        this.packageHandler = packageHandler;
        this.endpointService = endpointService;
        this.deploymentService = deploymentService;
        this.repositorySrv = repositorySrv;
        this.recoverySrv = recoverySrv;
        this.log = log;
        this.managerService = managerService;
    }

    /**
     * Process the deployment of the related service assembly. As we decide to
     * create xml task result only when deployment of the SA success, xml result
     * only handles xml component messages. All exceptions (caused by the
     * deployment framework or thrown by a component during the deployment of an
     * su) are thrown and they will stop the deployment process. It's a small
     * violation of the specification.
     * 
     * @param recovery
     *            true if called from the recovery service, false otherwise
     * @return xml management message handling the task result (see section 6.9
     *         of the JSR 208)
     * @throws Exception
     *             if an error occurs during the deployment proccess execution
     */
    public String deploySA(boolean recovery) throws Exception {

        URI installationRoot = null;

        /*
         * 0. Create framework task result. As we decide to create xml task
         * result only when deployment of the SA success, xml result only
         * handles xml component messages. All exceptions (caused by the
         * deployment framework or thrown by a component during the deployment
         * of an su) are thrown and they will stop the deployment process. It's
         * a small violation of the specification.
         */
        xmlResult = new XMLResult("deploy");
        createFrameworkTaskResult(true, getServiceAssemblyName());

        /*
         * 1. Check the validity of the service assembly package. Some checking
         * steps are described in the specification.
         */
        checkSAPackage();

        if (!recovery) {
            /*
             * 2.1 Expend service assembly package into the repository
             */
            installationRoot = explodeSA();
            /*
             * 2.2 Expend service unit packages into the repository
             */
            expendSUS(installationRoot);

        } else {
            installationRoot = repositorySrv.getServiceAssemblyInstallRoot(
                    getServiceAssemblyName()).toURI();
            retrieveSUInstallRoots(installationRoot);
        }

        /*
         * 3. Create service assembly lifecycle element
         */
        ServiceAssemblyLifeCycle saLifeCycle = createSALifeCycle(installationRoot);

        /*
         * 4. For each Service Unit included into the service assembly : execute
         * the deployment process of a service unit
         */
        deploySUS(installationRoot, saLifeCycle, recovery);

        /*
         * 5. Register connections if exist
         */
        registerConnections();

        /*
         * 6. Register Service assembly lifecycle into registry
         */
        registerSALifeCycle(saLifeCycle);

        File file = getArchiveFile(saArchiveURI);
        File installedArchive = getInstalledArchive(file);
        if (!recovery) {
            /*
             * 7. Creating state holder for service assembly
             */
            recoverySrv.createServiceAssemblyStateHolder(
                    getServiceAssemblyName(), repositorySrv
                            .getServiceAssemblyInstallRoot(
                                    getServiceAssemblyName()).toURL()
                            .toString(), installedArchive.toURL().toString(),
                    LifeCycleMBean.SHUTDOWN);

            /*
             * 8. Copy archive file to installed dir
             */
            copyFile(file, installedArchive);
        }

        /*
         * 9. Add sa name to the list of successfuly deployed sa
         */
        deploymentService.getDeployedSACache().add(getServiceAssemblyName());

        /*
         * 10. Create success file in workdir
         */
        createSuccessFile(installedArchive);

        return xmlResult.getString();
    }

    /**
     * Process the undeployment of the related service assembly.
     * 
     * @param saName
     *            the service assembly name
     * @return xml management message handling the task result (see section 6.9
     *         of the JSR 208)
     * @throws Exception
     *             if an error occurs during the undeployment proccess execution
     */
    public String undeploySA(String saName) throws Exception {
        String result = null;

        /*
         * 0. Create framework task result. As we decide to create xml task
         * result only when undeployment of the SA success, xml result only
         * handles xml component messages. All exceptions (caused by the
         * deployment framework or thrown by a component during the deployment
         * of an su) are thrown and they will stop the deployment process. It's
         * a small violation of the specification.
         */
        xmlResult = new XMLResult("undeploy");
        createFrameworkTaskResult(false, saName);

        /*
         * 1. Retrieve service assembly lifecycle
         */
        ServiceAssemblyLifeCycle saLifeCycle = retrieveServiceAssemblyLifeCycle(saName);

        /*
         * 2. Undeploy all SUs
         */
        undeploySUS(saLifeCycle);

        /*
         * 3. Remove possible connections
         */
        removeConnections(saLifeCycle);

        /*
         * 4. Remove the service assembly life cycle from the manager service SA
         * Map
         */
        managerService.unregisterServiceAssembly(saName);

        /*
         * 5. Remove service assembly from state holder file
         */
        recoverySrv.deleteServiceAssemblyStateHolder(saName);

        /*
         * 6. Remove service assembly from deployed sa list of deployment
         * service
         */
        deploymentService.getDeployedSACache().remove(saName);

        /*
         * 7. Remove the Service assembly folder.
         */
        saLifeCycle.removePackage();

        return result;
    }

    /**
     * Check the validity of connection elements
     * 
     * @param connections
     *            a {@link List} of {@link Connection} element
     * @throws Exception
     *             if consumer interface and consumer endpoint and consumer
     *             service are null or empty at the same time or if (provider
     *             endpoint and provider service) are null or empty
     */
    protected void checkConnections(List<Connection> connections)
        throws Exception {
        String msg = "";

        for (Connection connection : connections) {
            QName consInterface = connection.getConsumerInterfaceName();
            QName consService = connection.getConsumerServiceName();
            String consEndpoint = connection.getConsumerEndpointName();
            QName provService = connection.getProviderServiceName();
            String provEndpoint = connection.getProviderEndpointName();

            if (consInterface == null
                    || StringHelper.isEmpty(consInterface.getLocalPart())) {
                if (StringHelper.isEmpty(consEndpoint) || consService == null
                        || StringHelper.isEmpty(consService.getLocalPart())) {
                    msg = "Consumer interface and (consumer service or consumer endpoint) "
                            + "must not be null or empty at the same time.";
                    log.error(msg);
                    throw new Exception(msg);
                }
            }
            if (StringHelper.isEmpty(provEndpoint)
                    || (provService == null || StringHelper.isEmpty(provService
                            .getLocalPart()))) {
                msg = "Provider service and provider endpoint) "
                        + "must not be null or empty.";
                log.error(msg);
                throw new Exception(msg);
            }
        }
    }

    /**
     * Check the validity of the service assembly package. Some checking steps
     * are described in the specification (section 6.6 of the JSR 208).
     * 
     * @return true if checking is completly ok.
     * @throws Exception
     *             if one of the checking step fails
     */
    protected boolean checkSAPackage() throws Exception {

        String msg = null;

        /*
         * 0. Check if sa name is already used
         */
        if (isSANameAlreadyUsed()) {
            msg = "You are trying to deploy a sa with an identifier "
                    + "that is used by a previously successfuly deployed sa ("
                    + getServiceAssemblyName() + ").";
            log.error(msg);
            throw new Exception(msg);
        }

        /*
         * 1. Check service assembly identifier
         */
        if (StringHelper.isEmpty(getServiceAssemblyName())) {
            msg = "Unique identifier of a service assembly must be non null and non empty ("
                    + saArchiveURI.getPath() + ").";
            log.error(msg);
            throw new Exception(msg);
        }

        /*
         * 2. Check potential connection elements
         */
        List<Connection> connections = assembly.getConnections();
        if (connections != null) {
            checkConnections(connections);
        }

        /*
         * 3. Check potential service unit elements
         */
        List<ServiceUnit> sus = assembly.getServiceUnits();
        if (sus != null) {
            checkServiceUnits(sus);
        }

        return true;
    }

    /**
     * Check the validity of service unit elements
     * 
     * @param sus
     *            a {@link List} of {@link ServiceUnit} element
     * @throws Exception
     *             if su identifier is null or empty or non unique within the
     *             sa, if target artifact zip and target component are null or
     *             empty, if not all declared su zip are presents in the sa
     *             package or if su has already been deployed to the target
     *             component
     */
    protected void checkServiceUnits(List<ServiceUnit> sus) throws Exception {
        String msg = "";

        List<String> suIds = new ArrayList<String>();
        for (ServiceUnit su : sus) {
            String suIdentifier = su.getIdentification().getName();

            // Check su identifier : must be non null, non empty and unique
            // within the sa
            checkSUIdentifier(suIds, suIdentifier);

            suIds.add(suIdentifier);

            // check target artifact zip and target component entries : must
            // be non null and non empty
            String artifactZip = su.getTargetArtifactsZip();
            String targetComponent = su.getTargetComponentName();
            checkTargetComponentAndZip(artifactZip, suIdentifier,
                    targetComponent);

            // check that all declared su zip are presents in the sa package
            checkSUZipInSAZip(artifactZip);

            // Check if su hasn't already been deployed to the target
            // component
            if (isSUAlreadyDeployedOnTargetedComponent(su)) {
                msg = "You are trying to deploy a su ("
                        + su.getIdentification().getName()
                        + ")that have already been deployed on the targeted component ("
                        + su.getTargetComponentName() + ").";
                log.error(msg);
                throw new Exception(msg);
            }

        }
    }

    /**
     * Check su identifier : must be non null, non empty and unique within the
     * sa
     * 
     * @param suIds
     *            the su identifier list within the sa
     * @param suIdentifier
     *            the service unit identifier
     * @throws Exception
     *             if su identifier is null, empty and non unique within the sa
     */
    protected void checkSUIdentifier(List<String> suIds, String suIdentifier)
        throws Exception {
        String msg = "";
        if (StringHelper.isEmpty(suIdentifier)) {
            msg = "Identifier of a service unit must be non null and non empty (SA name : "
                    + getServiceAssemblyName() + ").";
            log.error(msg);
            throw new Exception(msg);
        }
        if (suIds.contains(suIdentifier)) {
            msg = "Identifier of a service unit must be unique within a single SA (SA name : "
                    + getServiceAssemblyName() + ").";
            log.error(msg);
            throw new Exception(msg);
        }
    }

    /**
     * Check that all declared su zip are presents in the sa package
     * 
     * @param artifactZip
     *            the artefact zip path into the sa archive
     * @throws Exception
     *             if at least one su is missing into the sa archive or if zip
     *             archive can't be correctly close or opened
     */
    protected void checkSUZipInSAZip(String artifactZip) throws Exception {
        String msg = "";
        ZipFile zipArchive = openZipFile(saArchiveURI);
        try {
            if (!hasEntry(zipArchive, artifactZip)) {
                msg = "The specified target artifact zip in JBI "
                        + "Descriptor's service unit declaration was not "
                        + "found within deployment package Zip archive file : "
                        + saArchiveURI.getPath() + ".";
                log.error(msg);
                throw new Exception(msg);
            }
        } finally {
            try {
                zipArchive.close();
            } catch (IOException e) {
                throw new Exception(
                        "Error while closing the Zip archive file.", e);
            }
        }
    }

    /**
     * Check target artifact zip and target component entries : must be non null
     * and non empty
     * 
     * @param artifactZip
     *            the artefact zip path into the sa archive
     * @param suIdentifier
     *            the service unit identifier
     * @param targetComponent
     *            the targeted component name
     * @throws Exception
     *             if target artifact zip or target component entries are null
     *             or empty
     */
    protected void checkTargetComponentAndZip(String artifactZip,
            String suIdentifier, String targetComponent) throws Exception {
        String msg = "";
        if (StringHelper.isEmpty(artifactZip)) {
            msg = "Target artifact zip of a service unit must be non null "
                    + "and non empty (SU name : " + suIdentifier + ").";
            log.error(msg);
            throw new Exception(msg);
        }
        if (StringHelper.isEmpty(targetComponent)) {
            msg = "Target component name of a service unit must be non null "
                    + "and non empty (SU name : " + suIdentifier + ").";
            log.error(msg);
            throw new Exception(msg);
        }
    }

    /**
     * Copy a file
     * 
     * @param src
     *            source file
     * @param dst
     *            destination file
     * @throws IOException
     *             if an error occurs during copy process
     */
    protected void copyFile(File src, File dst) throws IOException {
        FileUtils.copyFile(src, dst);
    }

    /**
     * Create the framework task result of the xml result message. Framework
     * task result is always a success because we decide that in other cases, an
     * exception is thrown by the SADeploymentManager.
     * 
     * @param isDeployment
     *            true in case of a deployment, false in case of an undeployment
     * @param saName
     *            the service assembly name
     */
    protected void createFrameworkTaskResult(boolean isDeployment, String saName) {
        if (isDeployment) {
            this.xmlResult.addFrameworkTaskResult(TaskResult.SUCCESS,
                    MessageType.INFO,
                    "Successfully deployed service assembly : {1}",
                    new String[] {saName}, CauseFramework.YES);
        } else {
            this.xmlResult.addFrameworkTaskResult(TaskResult.SUCCESS,
                    MessageType.INFO,
                    "Successfully undeployed service assembly : {1}",
                    new String[] {saName}, CauseFramework.YES);
        }
    }

    /**
     * Creates a new {@link ServiceAssemblyLifeCycle}
     * 
     * @param installationRoot
     *            the root {@link URI} of the service assembly into the
     *            repository
     * @return a new {@link ServiceAssemblyLifeCycle}
     * @throws ManagementException
     *             if deployment service MBean name retrieval fails
     */
    protected ServiceAssemblyLifeCycle createSALifeCycle(URI installationRoot)
        throws ManagementException {
        // Retrieve the DeploymentServiceImpl ObjectName
        ObjectName objectName = managerService.getDeploymentServiceMBeanName();

        return new ServiceAssemblyLifeCycle(objectName, log, assembly,
                repositorySrv, recoverySrv, installationRoot);

    }

    /**
     * Create a success file into the work directory to confirm the sa
     * successful deployment
     * 
     * @param installedArchive
     *            the installed archive {@link File}
     * @return true if succes file creation is ok, false otherwise
     * @throws IOException
     *             if success file creation fails
     */
    protected boolean createSuccessFile(File installedArchive)
        throws IOException {
        return new File(AutoLoaderImpl.getWorkDirectory(), installedArchive
                .getName()
                + ".success").createNewFile();
    }

    /**
     * * Deploy all Service Units to the JBI container
     * 
     * @param installationRoot
     *            the root {@link URI} of the service assembly
     * @param saLifeCycle
     *            the {@link ServiceAssemblyLifeCycle} of the service assembly
     * @param recovery
     *            true if called from the recovery service, false otherwise
     * @throws ManagementException
     *             if an sus deployment fails caused by a severe error of the
     *             container
     * 
     */
    protected void deploySUS(URI installationRoot,
            ServiceAssemblyLifeCycle saLifeCycle, boolean recovery)
        throws ManagementException {
        List<ServiceUnit> serviceUnits = assembly.getServiceUnits();
        if (serviceUnits != null) {
            StringBuffer componentOutput = new StringBuffer();
            for (ServiceUnit serviceUnit : serviceUnits) {
                SUDeploymentManager manager = new SUDeploymentManager(
                        serviceUnit, suInstallRoots.get(serviceUnit
                                .getIdentification().getName()), saLifeCycle,
                        repositorySrv, managerService, log, packageHandler,
                        deploymentService);
                String xmlReturn = manager.deploySU(recovery);
                if (!isValidComponentTask(xmlReturn, this.getClass()
                        .getClassLoader().getResourceAsStream(
                                "schema/management-message-component.xsd"))) {
                    xmlReturn = xmlResult.wrapComponentTaskResult(serviceUnit
                            .getTargetComponentName(), "deploy", xmlReturn);
                }
                componentOutput.append(xmlReturn);
            }
            this.xmlResult.addComponentTaskResult(componentOutput);
        }
    }

    /**
     * Expand a service unit package into its service assembly installRoot.
     * 
     * @param suId
     *            the unique identifier of the su
     * @param suZipLocation
     *            URI representing the su zip location
     * @param saId
     *            the unique identifier of the sa
     * @return SU installationRoot, null if it fails
     * @throws ManagementException
     *             if it fails to expend the su package to the repository
     */
    protected URI expandSUIntoSA(String suId, URI suZipLocation, String saId)
        throws ManagementException {
        log.start();
        String msg;
        URI result = null;

        try {
            result = repositorySrv.explodeSUIntoSAInstallFolder(suId,
                    suZipLocation, saId);
        } catch (PetalsException pe) {
            msg = "Unexpected error expending service JBI installation "
                    + "package into repository.";
            log.error(msg, pe);
            throw new ManagementException(msg, pe);
        }
        if (result == null) {
            msg = "Unexpected platform repository service error. The "
                    + "su root must NOT be null: " + suId;
            log.error(msg);
            throw new ManagementException(msg);
        }
        log.end();
        return result;
    }

    /**
     * Explode service unit packages into the repository *
     * 
     * @param saInstallRoot
     *            the root {@link URI} of the service assembly
     * @throws ManagementException
     *             if the repository service doesn't manage to expend su archive
     */
    protected void expendSUS(URI saInstallRoot) throws ManagementException {
        List<ServiceUnit> sus = assembly.getServiceUnits();
        for (ServiceUnit unit : sus) {
            URI suArchiveURI = getSUArchiveURI(unit, saInstallRoot);
            URI suInstallRoot = expandSUIntoSA(unit.getIdentification()
                    .getName(), suArchiveURI, getServiceAssemblyName());
            suInstallRoots.put(unit.getIdentification().getName(),
                    suInstallRoot);
        }
    }

    /**
     * Expend service assembly package into the repository
     * 
     * @return return installation {@link URI} of the expanded package, null if
     *         it fails
     * @throws ManagementException
     *             when it fails to expand service assembly package into
     *             repository
     */
    protected URI explodeSA() throws ManagementException {
        return packageHandler.expandPackageIntoRepository(
                getServiceAssemblyName(), saArchiveURI, EntityType.SA_TYPE);
    }

    /**
     * Get the file according to the archive URI
     * 
     * @param archiveURI
     *            URI of the archive
     * @return a File pointing on the archive
     */
    protected File getArchiveFile(URI archiveURI) {
        return new File(archiveURI);
    }

    /**
     * Get a file of the given name in the "installed directory"
     * 
     * @param file
     *            the file to look for
     * @return a File pointing on the file name in the "installed directory"
     */
    protected File getInstalledArchive(File file) {
        return new File(AutoLoaderImpl.getInstalledDirectory(), file.getName());
    }

    /**
     * Retrieve the service assembly name found in the SA production element
     * 
     * @return the service assembly name
     */
    protected String getServiceAssemblyName() {
        return assembly.getIdentification().getName();
    }

    /**
     * Retrieve the service unit archive {@link URI}
     * 
     * @param su
     *            the {@link ServiceUnit} production element
     * @param saInstallRoot
     *            the sa installation root {@link URI}
     * @return the su archive {@link URI}
     */
    protected URI getSUArchiveURI(ServiceUnit su, URI saInstallRoot) {
        String suZipNAme = su.getTargetArtifactsZip();
        String suZipURL = "file://" + saInstallRoot.getPath() + suZipNAme;
        return packageHandler.processAndGetPackageURI(suZipURL.replaceAll(" ",
                "%20"), true);
    }

    /**
     * Test if a zip file has a specific entry
     * 
     * @param archive
     *            the sa {@link ZipFile}
     * @param entryName
     *            the entry name
     * @return true if the entry is in the zip file
     */
    protected boolean hasEntry(ZipFile zipArchive, String artifactZip) {
        return ZipUtil.hasEntry(zipArchive, artifactZip);
    }

    /**
     * Check if the sa name is already used by a previously successfuly deployed
     * sa.
     * 
     * @return true if the sa name is already used, false otherwise
     */
    protected boolean isSANameAlreadyUsed() {

        List<String> deployedSACache = deploymentService.getDeployedSACache();

        return deployedSACache.contains(getServiceAssemblyName());
    }

    /**
     * Check if the su is already deployed on the targeted component.
     * 
     * @param su
     *            the service unit production element
     * 
     * @return true if the su is already deployed on the targeted component,
     *         false otherwise
     * @throws Exception
     *             if deploymentService.isDeployedServiceUnit fails
     */
    protected boolean isSUAlreadyDeployedOnTargetedComponent(ServiceUnit su)
        throws Exception {
        return deploymentService.isDeployedServiceUnit(su
                .getTargetComponentName(), su.getIdentification().getName());
    }

    /**
     * Check if a component task result is valid against an xml schema
     * 
     * @param componentTask
     *            xml of the component task
     * @param xsd
     *            an {@link InputStream} of the xsd file
     * @return true if valid, false if not valid
     */
    protected boolean isValidComponentTask(String componentTask, InputStream xsd) {
        return XMLResult.isValidComponentTask(componentTask, xsd);
    }

    /**
     * Load a {@link ZipFile} from an {@link URI}
     * 
     * @param archiveURI
     *            {@link URI} of the {@link ZipFile}
     * @return the targeted {@link ZipFile}
     * @throws PetalsException
     *             when {@link ZipFile} can't be opened
     */
    protected ZipFile openZipFile(URI archiveURI) throws PetalsException {
        return ZipUtil.openZipFile(archiveURI);
    }

    /**
     * Register a connection into the endpoint registry of the JBI container
     * 
     * @param connection
     *            The {@link Connection} to register
     * @throws ManagementException
     *             when container fails to register a connection into the
     *             endpoint registry
     */
    protected void registerConnection(Connection connection)
        throws ManagementException {
        String saName = getServiceAssemblyName();

        QName consInterface = connection.getConsumerInterfaceName();
        QName consService = connection.getConsumerServiceName();
        String consEndpoint = connection.getConsumerEndpointName();
        QName provService = connection.getProviderServiceName();
        String provEndpoint = connection.getProviderEndpointName();

        try {
            if (consInterface != null) {

                endpointService.createConnection(consInterface, provService,
                        provEndpoint);

            } else {
                endpointService.createConnection(consService, consEndpoint,
                        provService, provEndpoint);
            }
        } catch (JBIException e) {
            String msg = "Error while registering a service unit connection for the service assembly: "
                    + saName + ".";
            log.error(msg);
            throw new ManagementException(msg, e);
        }

    }

    /**
     * Register connections, if exist, into Registry
     * 
     * @throws ManagementException
     *             when container fails to register a particular connection
     * 
     */
    protected void registerConnections() throws ManagementException {
        List<Connection> connections = assembly.getConnections();
        if (connections != null) {
            for (Connection connection : connections) {
                registerConnection(connection);
            }
        }
    }

    /**
     * Register ServiceAssemblyLifeCycle into manager service map.
     * 
     * @param saLifeCycle
     *            the {@link ServiceAssemblyLifeCycle} to register
     */
    protected void registerSALifeCycle(ServiceAssemblyLifeCycle saLifeCycle) {
        managerService.registerServiceAssembly(getServiceAssemblyName(),
                saLifeCycle);

    }

    /**
     * Remove all connections registered during the SA deployment process
     * 
     * @param saLifeCycle
     *            the service assembly life cycle
     * @throws Exception
     *             if problem occurs during the deletion of a specific
     *             connection
     */
    protected void removeConnections(ServiceAssemblyLifeCycle saLifeCycle)
        throws Exception {
        ServiceAssembly saDesc = saLifeCycle.getServiceAssembly();
        List<Connection> connections = saDesc.getConnections();
        if (connections != null) {
            for (Connection connection : connections) {
                QName consInterface = connection.getConsumerInterfaceName();
                QName consService = connection.getConsumerServiceName();
                String consEndpoint = connection.getConsumerEndpointName();
                QName provService = connection.getProviderServiceName();
                String provEndpoint = connection.getProviderEndpointName();

                if (provEndpoint != null && provService != null) {
                    if (consInterface != null) {
                        endpointService.deleteConnection(consInterface,
                                provService, provEndpoint);
                    } else if (consService != null && consEndpoint != null) {
                        endpointService.deleteConnection(consService,
                                consEndpoint, provService, provEndpoint);
                    }
                }
            }
        }
    }

    /**
     * Retrieve the {@link ServiceAssemblyLifeCycle} matching the given service
     * assembly name.
     * 
     * @param saName
     *            the service assembly name
     * @return the retrieved {@link ServiceAssemblyLifeCycle}
     * @throws Exception
     *             if no service assembly can be retrieved
     */
    protected ServiceAssemblyLifeCycle retrieveServiceAssemblyLifeCycle(
            String saName) throws Exception {
        ServiceAssemblyLifeCycle saLifeCycle = managerService
                .getServiceAssemblyByName(saName);
        if (saLifeCycle == null) {
            String msg = "The service assembly (" + saName
                    + ") isn't deployed into the JBI environment.";
            log.error(msg);
            throw new Exception(msg);
        }
        return saLifeCycle;
    }

    /**
     * Retrieve all su installation root {@link URI}
     * 
     * @param installationRoot
     *            the sa intallation root {@link URI}
     */
    protected void retrieveSUInstallRoots(URI installationRoot) {
        List<ServiceUnit> sus = assembly.getServiceUnits();
        for (ServiceUnit unit : sus) {
            String suName = unit.getIdentification().getName();
            suInstallRoots.put(suName, packageHandler.processAndGetPackageURI(
                    "file://" + installationRoot.getPath() + suName, true));
        }

    }

    /**
     * Undeploy all SUs by calling the "undeploySU" method of the corresponding
     * SUDeploymentManager.
     * 
     * @param saLifeCycle
     *            the parent {@link ServiceAssemblyLifeCycle}
     * @throws Exception
     *             if an error occurs during the undeployment of a service unit
     */
    protected void undeploySUS(ServiceAssemblyLifeCycle saLifeCycle)
        throws Exception {
        /*
         * Put the service assembly in the shutdown state
         */
        String saState = saLifeCycle.getCurrentState();
        if (LifeCycleAbstract.STARTED.equals(saState)) {
            saLifeCycle.stop();
            saLifeCycle.shutDown();
        } else if (LifeCycleAbstract.STOPPED.equals(saState)) {
            saLifeCycle.shutDown();
        }
        saState = saLifeCycle.getCurrentState();
        if (saState.equals(LifeCycleAbstract.SHUTDOWN)) {
            /*
             * The service assembly is in the SHUTDOWN state We can try to
             * undeploy it
             */

            /*
             * Jbi spec specifies that service units must be undeployed in the
             * reverse order of the declared list in service assembly descriptor
             */
            List<ServiceUnit> suList = saLifeCycle.getServiceAssembly()
                    .getServiceUnits();
            Collections.reverse(suList);
            StringBuffer componentOutput = new StringBuffer();
            for (ServiceUnit suDesc : suList) {
                SUDeploymentManager manager = new SUDeploymentManager(suDesc,
                        null, saLifeCycle, repositorySrv, managerService, log,
                        packageHandler, deploymentService);
                String xmlReturn = manager.undeploySU();
                if (!isValidComponentTask(xmlReturn, this.getClass()
                        .getClassLoader().getResourceAsStream(
                                "schema/management-message-component.xsd"))) {
                    xmlReturn = xmlResult.wrapComponentTaskResult(suDesc
                            .getTargetComponentName(), "undeploy", xmlReturn);
                }
                componentOutput.append(xmlReturn);
            }
            this.xmlResult.addComponentTaskResult(componentOutput);
        }
    }

}
