/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramTransporter.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.transport;

import static org.objectweb.petals.util.InterfaceNamesHelper.JBI_ROUTER_NAME;
import static org.objectweb.petals.util.InterfaceNamesHelper.LOGGER_NAME;

import javax.jbi.JBIException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.naming.InitialContext;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.jbi.routing.RoutingException;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.petals.util.SystemUtil;

/**
 * Implementation of the <code>Transporter</code> with JORAM. <br>
 * This implementation uses the fractal JoramAgent. <br>
 * Each <code>JoramTransporter</code> registrers a <code>Topic</code> in the
 * Joram distributed JNDI with the following name :
 * <em>JoramConnection-[containerId]</em> <br>
 * The name of a container is the <em>ID</em> specified in the
 * <em>conf/server.properties</em> file. <br>
 * This <code>JoramTransporter</code> uses an <code>ObjectSerializer</code>
 * to create <em>jms.ObjectMessage</em> to send via JMS. So, the
 * <code>MessageExchange</code> to send is serialized.
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 17:18:07 +0200 (lun., 29 mai 2006) $
 * @since Petals 1.0
 * @author alouis
 */
public class JoramTransporter implements Transporter, BindingController,
    LifeCycleController {

    private static final String[] CLIENT_INTERFACES = new String[] {JBI_ROUTER_NAME};

    /**
     * Component LifeCycle state.
     */
    protected String fcState;

    protected String host;

    protected int id;

    protected JoramConnection joramConnection;

    protected Serializer serializer = new ObjectSerializer();

    /**
     * Logger
     */
    protected LoggingUtil log;

    protected String pwd;

    protected Router router;

    protected int tcp;

    protected String user;

    /**
     * Send the message to the specified container. The send can be local or
     * distant
     */
    public void send(MessageExchangeImpl messageExchange, String containerName,
        long timeToLive) throws TransportException {

        if (id == Integer.parseInt(containerName)) {
            sendLocal(messageExchange);
        } else {
            sendDistant(messageExchange, containerName, timeToLive);

        }
    }

    /**
     * init variables with joram.properties file and create/bind the topic for
     * this container. A listener is created and started to listen incoming
     * messages
     * 
     * @throws Exception
     */
    protected void startJMS() throws Exception {
        log.start();
        // id = Integer.parseInt(PropertyUtil.getProperty(
        // PropertyUtil.SERVER_PROPS, "joram.id"));
        id = Integer.parseInt(SystemUtil.getContainerName());

        tcp = Integer.parseInt(SystemUtil.getJoramTCPPort());
        user = SystemUtil.getJoramUser();
        pwd = SystemUtil.getJoramPassword();
        host = SystemUtil.getHost();

        // create initialContext
        InitialContext context = new javax.naming.InitialContext(PropertyUtil
            .retrieveJNDIProperties());

        joramConnection = new JoramConnection(id, tcp, user, pwd, host, context);

        log.end();
    }

    public void startListening() throws TransportException {
        log.start();

        MessageListener aML = new MessageListener() {
            public void onMessage(Message aMessage) {
                processIncomingMessage(aMessage);
            }
        };
        try {
            joramConnection.start(aML);
        } catch (Exception e) {
            throw new TransportException(e);
        }

        log.end();
    }

    /**
     * TODO manage exceptions This method is called asynchronously when a
     * message is received from Joram, via the topic listener.
     * 
     * The recevied message is transform into a <code>MessageExchangeImpl</code>
     * by using the <code>Serializer</code>, and is sent to the
     * <code>Router</code> by calling <code>receive()</code>.
     * 
     * @param msg
     *            the jms message TODO processIncomingMessage :manage the
     *            reception of unknown message type. PB between the calculation
     *            of the next recipient and notion of service provider/consumer
     */
    protected void processIncomingMessage(Message msg) {
        log.start();
        if (msg instanceof ObjectMessage) {
            try {
                MessageExchangeImpl jbiMsg = serializer.jms2jbi(msg);

                // transfer the message to the router
                router.receive(jbiMsg);
                
                //ack the message
                msg.acknowledge();
            } catch (Exception e) {
                log.error(e);
            }
        } else {
            log.error("the received JMS message is not an objectmessage");
        }
        log.end();
    }

    /**
     * Send the message to the specified distant container, by serializing the
     * exhange with a <code>Serializer</code>. We retrieve the related
     * distant <code>Topic</code> with the namle of the container, transform
     * the <code>MessageExchangeImpl</code> into a <code>jms.Message</code>,
     * and send this message onto the distant topic.
     * 
     * @param messageExchange
     *            must be non null
     * @param containerName
     *            must be non null
     * @param timeToLive
     * @throws TransportException
     *             container not found, invalid message.
     */
    protected void sendDistant(MessageExchangeImpl messageExchange,
        String containerName, long timeToLive) throws TransportException {
        log.start();
        try {
            joramConnection.sendTo(containerName, messageExchange, timeToLive);
        } catch (TransportException e) {
            log.error(e.getMessage(),e);
            throw e;
        }
        log.end();
    }

    /**
     * Send the message locally. No serialization of the message
     * 
     * @param messageExchange
     *            a messageExchange, can be null
     * @throws TransportException
     *             problem with the <code>Router</code>
     */
    protected void sendLocal(MessageExchangeImpl messageExchange)
        throws TransportException {
        log.start();
        try {
            router.receive(messageExchange);
        } catch (RoutingException e) {
            String msg = "Can not send locally the MessageExchange to the local Router.";
            log.error(msg,e);
            throw new TransportException(msg,e);
        }
        log.end();
    }

    /**
     * close the JMS connection (but does not close the joram server)
     * 
     * @throws JBIException
     */
    protected void stopJMS() throws TransportException {
        log.start();
        try {
            joramConnection.stop();
        } catch (Exception e) {
            String msg = "Can not stop the Transporter.";
            log.error(msg,e);
            throw new TransportException(msg,e);
        }
        log.end();
    }


    /**
     * Shutdown the Transporter.
     * The related Topic is removed from Joram
     * and from the distributed Jndi server 
     */
    public void shutdown() throws TransportException {
        try {
            joramConnection.unsubscribe();
        } catch (Exception e) {
            String msg = "Can not shutdown correctly the Transporter.";
            log.error(msg,e);
            throw new TransportException(msg,e);
        }
    }

    
    // ----------------------------------------------------------------------
    // Fractal BindingController implementation
    // ----------------------------------------------------------------------


    /**
     * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
     *      java.lang.Object)
     */
    public void bindFc(String clientItfName, Object serverItf)
        throws NoSuchInterfaceException, IllegalBindingException,
        IllegalLifeCycleException {
        if (LOGGER_NAME.equals(clientItfName)) {
            log = new LoggingUtil((Logger) serverItf);
        } else if (JBI_ROUTER_NAME.equals(clientItfName)) {
            router = (Router) serverItf;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
     */
    public String getFcState() {
        return STARTED.equals(fcState) ? STARTED : STOPPED;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#listFc()
     */
    public String[] listFc() {
        return CLIENT_INTERFACES;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
     */
    public Object lookupFc(String clientItfName)
        throws NoSuchInterfaceException {
        if (JBI_ROUTER_NAME.equals(clientItfName)) {
            return router;
        } else if (LOGGER_NAME.equals(clientItfName)) {
            return log;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    public void startFc() throws IllegalLifeCycleException {
        try {
            startJMS();
            fcState = STARTED;
        } catch (Exception e) {
            log.error(e);
        }
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
     */
    public void stopFc() throws IllegalLifeCycleException {
        // FIXME problem because the JNDI directory is already stopped
        try {
            stopJMS();
        } catch (JBIException e) {
            log.error("Error stopping JMS.", e);
        }
        fcState = STOPPED;
    }

    /**
     * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
     */
    public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
        IllegalBindingException, IllegalLifeCycleException {
        if (LOGGER_NAME.equals(clientItfName)) {
            log = null;
        } else if (JBI_ROUTER_NAME.equals(clientItfName)) {
            router = null;
        } else {
            throw new NoSuchInterfaceException(clientItfName);
        }
    }

}