package org.objectweb.petals.tools.jbicommon.descriptor;

import javax.xml.namespace.QName;

/**
 * The consumes element is used to declare a service consumed by a service unit.
 * A consumed service must be define with the following required attribute:
 * <ul>
 * <li><b>interface-name.</b> This qualified name indicates the type of
 * service to be consumed.</li>
 * </ul>
 * Optionally a fully qualified name of the endpoint the SU will use to consume
 * the service can be specified. If the service endpoint name is declared, an
 * optional attribute can be supplied:
 * <ul>
 * <li><b>link-type.</b> This attribute is used to indicate the expected use
 * of this consumer with connection links. The possible values of this
 * attributes are:</li>
 * <ul>
 * <li><b>standard.</b> This is the default value. This indicates that the
 * provided service endpoint must be routed according to standard normalized
 * message routing rules.</li>
 * <li><b>hard.</b>This indicates that the provided service endpoint name must
 * match a service provider's service endpoint name; indirect connections are
 * not allowed.</li>
 * <li><b>soft.</b>This indicates that the provided service endpoint must
 * <i>not</i> match a service provider's service endpoint name; rather, it must
 * match an indirect connection name.</li>
 * </ul>
 * </ul>
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 15:18:07Z $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class Consumes {

    /**
     * Standard Link Type.
     */
    public static final String STANDARD_LINK_TYPE = "standard";

    /**
     * Hard Link Type.
     */
    public static final String HARD_LINK_TYPE = "hard";

    /**
     * Soft Link Type.
     */
    public static final String SOFT_LINK_TYPE = "soft";

    /**
     * Qualified interface name.
     */
    private QName interfaceName;

    /**
     * Service name to be provided.
     */
    private QName serviceName;

    /**
     * Endpoint name to be provided.
     */
    private String endpointName;

    /**
     * The optional link type.
     */
    private String linkType;

    /**
     * Default constructor.
     */
    public Consumes() {
        super();
    }

    //
    // Getters and Setters
    //

    /**
     * Returns the optional fully qualified endpoint which the service unit will
     * use to consume the service.
     * 
     * @return <code>String</code> the endpoint name.
     */
    public String getEndpointName() {
        return endpointName;
    }

    /**
     * Return the qualified name indicating the type of service to be consumed.
     * 
     * @return <code>QName</code> the interface name.
     */
    public QName getInterfaceName() {
        return interfaceName;
    }

    /**
     * Returns the optional fully qualified service name which the service unit
     * will use to consume the service.
     * 
     * @return <code>QName</code> the service name.
     */
    public QName getServiceName() {
        return serviceName;
    }

    /**
     * Set the optional fully qualified endpoint name which the service unit
     * will consume.
     * 
     * @param endpointName
     *            the endpoint name.
     */
    void setEndpointName(String endpointName) {
        this.endpointName = endpointName;
    }

    /**
     * Set the interface name indicating the type of service to be consumed.
     * 
     * @param interfaceName
     *            the interface name.
     */
    void setInterfaceName(QName interfaceName) {
        this.interfaceName = interfaceName;
    }

    /**
     * Set the optional fully qualified service name which the service unit will
     * consume.
     * 
     * @param serviceName
     *            the service name.
     */
    void setServiceName(QName serviceName) {
        this.serviceName = serviceName;
    }

    /**
     * Returns the optional link type associated to a fully qualified endpoint
     * the SU will use to consume the service.
     * 
     * @return <code>String</code> the link type: <code>"standard"</code>,
     *         <code>"hard"</code> or <code>"soft"</code>.
     */
    public String getLinkType() {
        return linkType;
    }

    /**
     * Set the optional link type.
     * 
     * @param type
     *            the link type.
     */
    void setLinkType(String type) {
        this.linkType = type;
    }

    //
    // implementation methods
    //

    /**
     * Returns <code>true</code> if the standard link type is used.
     */
    public boolean isStandardLink() {
        if (hasQualifiedEndpoint()) {
            return linkType == null || linkType.equals(STANDARD_LINK_TYPE);
        } else {
            return false;
        }
    }

    /**
     * Returns <code>true</code> if the hard link type is used.
     */
    public boolean isHardLink() {
        if (hasQualifiedEndpoint()) {
            return linkType != null && linkType.equals(HARD_LINK_TYPE);
        } else {
            return false;
        }
    }

    /**
     * Returns <code>true</code> if the soft link type is used.
     */
    public boolean isSoftLink() {
        if (hasQualifiedEndpoint()) {
            return linkType != null && linkType.equals(SOFT_LINK_TYPE);
        } else {
            return false;
        }
    }

    /**
     * Returns <code>true</code> if the optional fully qualified endpoint name
     * have been specified. Otherwise, returns <code>false</code>.
     */
    boolean hasQualifiedEndpoint() {
        return serviceName != null && endpointName != null;
    }
}