/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JBIDescriptorBuilder.java 565 2006-06-08 10:03:27Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.jbicommon.descriptor;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.objectweb.petals.tools.jbicommon.util.FileUtil;
import org.objectweb.petals.tools.jbicommon.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import com.sun.org.apache.xpath.internal.XPathAPI;

/**
 * TODO refactor algorithm The JBIDescriptorBuilder helps to populate a
 * <code>JBIDescriptor</code> instance with values loaded from a specified
 * <i>jbi.xml</i>.
 * 
 * @version $Rev: 565 $ $Date: 2006-06-08 10:03:27Z $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class JBIDescriptorBuilder {

    /**
     * This class is used to handle errors during creating XML objects, parsing
     * XML and validating a XML.
     */
    protected static class JbiXmlErrorHandler implements ErrorHandler {

        private Logger log;

        /**
         * 
         */
        public JbiXmlErrorHandler(Logger loggingUtil) {
            super();
            log = loggingUtil;
        }

        /**
         * @see org.xml.sax.ErrorHandler#error(org.xml.sax.SAXParseException)
         */
        public void error(SAXParseException se) throws SAXException {
            log.log(Level.SEVERE, se.getMessage(), se);
            throw se;
        }

        /**
         * @see org.xml.sax.ErrorHandler#fatalError(org.xml.sax.SAXParseException)
         */
        public void fatalError(SAXParseException se) throws SAXException {
            log.log(Level.SEVERE, se.getMessage(), se);
            throw se;
        }

        /**
         * @see org.xml.sax.ErrorHandler#warning(org.xml.sax.SAXParseException)
         */
        public void warning(SAXParseException e) throws SAXException {
            log.log(Level.WARNING, e.getMessage(), e);
        }

    }

    /**
     * Standard location to the JBI Descriptor resource.
     */
    public static final String JBI_DESCRIPTOR_RESOURCE = "META-INF"
        + File.separator + "jbi.xml";

    /**
     * The JBI Descriptor XML Schema Resource Location ("schema/jbi.xsd").
     */
    public static final String JBI_SCHEMA_RESOURCE = "schema"
        + File.separator + "jbi.xsd";

    /**
     * URI to the jbi.xml file.
     */
    protected URI jbiXmlUri;

    /**
     * Logger wrapper
     */
    protected Logger log;

    /**
     * Schema file
     */
    protected Source schemaFile;

    /**
     * The source XML document file.
     */
    protected Document source;

    /**
     * JBI Descriptor Builder constructor using the source xml file URI from
     * which the document will be loaded.
     */
    public JBIDescriptorBuilder(URI sourceUri, Logger logger) {
        if (sourceUri == null) {
            throw new IllegalArgumentException(
                "DescriptorBuilder: invalid source URI argument.");
        }

        this.jbiXmlUri = sourceUri;
        log = logger;
    }

    // ----------------------------------------------------------------------
    // Getters and Setters
    // ----------------------------------------------------------------------

    /**
     * Load, Parse and Validate jbi.xml document for building (populating) the
     * <code>Descriptor</code> object.
     * 
     * @throws PetalsException
     *             unexpected exception.
     * @throws SAXException
     *             in case a XML validation error occur.
     */
    public JBIDescriptor build() throws SAXException {

        // do not repeat load, parse and validation steps
        if (source == null) {

            // validate the jbi.xml file against the schema
            try {
                // FIXME we do not check if the jbi.xml is valid
                validateJbiXml(jbiXmlUri);

                // parse XML after validation
                source = parseJbiXml(jbiXmlUri);

            } catch (IOException ioe) {
                log.log(Level.SEVERE, ioe.getMessage(), ioe);

                throw new SAXException(
                    "Unexpected I/O exception using resource: "
                        + jbiXmlUri.toString(), ioe);
            }
        }
        return buildDescriptor(source);
    }

    /**
     * Returns a String representation of a JBI descriptor file
     * 
     * @param descriptorFile
     *            the JBI Descriptor <code>File</code>
     * @return a <code>String</code> representation of the JBI Descriptor
     */
    public String getJbiDescriptorAsString() {
        File descriptorFile = new File(jbiXmlUri);
        if (descriptorFile.isFile()) {
            ByteArrayOutputStream os = null;
            InputStream is = null;
            try {
                os = new ByteArrayOutputStream();
                is = new FileInputStream(descriptorFile);
                FileUtil.copyInputStream(is, os);
                return os.toString();
            } catch (Exception e) {
                log.log(Level.SEVERE, "Error reading jbi descritor: "
                    + descriptorFile, e);
            }
        }
        return null;
    }

    // ----------------------------------------------------------------------
    // implementation methods
    // ----------------------------------------------------------------------

    /**
     * @return Returns the source.
     */
    public Document getSource() {
        return source;
    }

    /**
     * @return Returns the jbiXmlUri.
     */
    public URI getURI() {
        return jbiXmlUri;
    }

    /**
     * 
     * @param sourceURI
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public Document parseJbiXml(URI sourceURI) throws SAXException, IOException {

        DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
            .newInstance();

        DocumentBuilder docBuilder = null;

        try {
            docBuilder = docBuilderFactory.newDocumentBuilder();
        } catch (ParserConfigurationException pce) {
            log.log(Level.SEVERE, pce.getMessage(), pce);
            return null;
        }

        // set up xml parsing error handler
        JbiXmlErrorHandler xmlErrorHandler = new JbiXmlErrorHandler(log);
        docBuilder.setErrorHandler(xmlErrorHandler);

        // parse the xml document URI to return
        String jbiXmlURI = sourceURI.toString();

        Document document = null;

        try {
            document = docBuilder.parse(jbiXmlURI);
        } catch (SAXException saxe) {
            log.log(Level.SEVERE, saxe.getMessage(), saxe);
        } catch (IOException ioe) {
            log.log(Level.SEVERE, ioe.getMessage(), ioe);
        }
        return document;
    }

    /**
     * 
     * @param sourceURI
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public boolean validateJbiXml(URI sourceURI) throws SAXException,
        IOException {

        if (sourceURI == null) {
            return false;
        }

        File document = new File(sourceURI);

        // instance for xml error handling
        JbiXmlErrorHandler xmlErrorHandler = new JbiXmlErrorHandler(log);

        // create a SchemaFactory capable of understanding WXS schemas
        SchemaFactory factory = SchemaFactory
            .newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

        factory.setErrorHandler(xmlErrorHandler);

        // load a WXS schema, represented by a Schema instance
        if (schemaFile == null) {
            schemaFile = new StreamSource(openJbiSchemaFile());
        }
        Schema schema;

        try {
            schema = factory.newSchema(schemaFile);
        } catch (SAXException saxe) {
            log.log(Level.SEVERE, saxe.getMessage(), saxe);
            return false;
        }

        // create a Validator instance which is be used to validate a document
        Validator validator = schema.newValidator();
        validator.setErrorHandler(xmlErrorHandler);

        // validate the DOM tree
        try {
            validator.validate(new StreamSource(document));
        } catch (SAXParseException spe) {
            // error thrown by error handler
            return false;
        } catch (SAXException saxe) {
            log.log(Level.SEVERE, saxe.getMessage(), saxe);
            throw saxe;
        } catch (IOException ioe) {
            log.log(Level.SEVERE, ioe.getMessage(), ioe);
            throw ioe;
        }

        // succesfully validated without errors
        return true;
    }

    /**
     * Parse the XML document and load values to the Descriptor object
     * representation.
     * 
     * @param document
     * @return a populated Descriptor instance with values loaded from the XML
     *         file. May return <code>null</code> if an error occur while
     *         navigating on XML tree.
     */
    protected JBIDescriptor buildDescriptor(Document document) {

        JBIDescriptor descriptor = new JBIDescriptor();

        Node productionTree;

        try {
            // get root element
            Element jbi = document.getDocumentElement();

            // extract JBI version attribute from the root element
            String versionAttr = jbi.getAttribute("version");
            descriptor.setVersion(Double.parseDouble(versionAttr));

            // component production
            productionTree = XPathAPI.selectSingleNode(document,
                "jbi/component");
            if (productionTree != null) {
                descriptor.setComponent(loadComponentXml(document,
                    productionTree));
            } else if (XPathAPI
                .selectSingleNode(document, "jbi/shared-library") != null) {
                // shared library production
                descriptor.setSharedLibrary(loadSharedLibraryXml(XPathAPI
                    .selectSingleNode(document, "jbi/shared-library")));
            } else if (XPathAPI.selectSingleNode(document,
                "jbi/service-assembly") != null) {
                // service assembly production
                descriptor.setServiceAssembly(loadServiceAssemblyXml(XPathAPI
                    .selectSingleNode(document, "jbi/service-assembly")));
            } else if (XPathAPI.selectSingleNode(document, "jbi/services") != null) {
                // services production
                descriptor.setServices(loadServicesXml(XPathAPI
                    .selectSingleNode(document, "jbi/services")));
            } else {
                // no production found
                return null;
            }

        } catch (TransformerException te) {
            log.log(Level.SEVERE, te.getMessage(), te);
            return null;
        }

        return descriptor;
    }

    /**
     * @param node
     * @param attrName
     * @return
     */
    protected QName extractXmlAttributeQName(Node node, String attrName) {
        String attr = XMLUtil.getAttributeValue(node, attrName);

        QName qName = null;

        // qname like "xmlns:name"
        if ((attr.indexOf(":") > -1)
            && (!attr.startsWith("{"))) {
            String ns = attr.substring(0, attr.indexOf(":"));

            String namespace = XMLUtil
                .getNamespaceURIForPrefix(node, ns, false);
            /*
             * String namespace = source.getDocumentElement().getAttribute(
             * "xmlns:" + ns);
             */

            qName = new QName(namespace, attr.substring(attr.indexOf(":") + 1));
        } else {
            // qname like "{ns}name" or "name"
            qName = QName.valueOf(attr);
        }
        return qName;
    }

    // ----------------------------------------------------------------------
    // private static methods
    // ----------------------------------------------------------------------

    /**
     * @param element
     * @throws TransformerException
     */
    protected ComponentDescription loadComponentXml(Document doc,
        Node production) throws TransformerException {
        ComponentDescription component = new ComponentDescription();

        // component attributes
        component.setType(XMLUtil.getAttributeValue(production, "type"));
        component.setComponentClassLoaderDelegation(XMLUtil.getAttributeValue(
            production, "component-class-loader-delegation"));
        component.setBootstrapClassLoaderDelegation(XMLUtil.getAttributeValue(
            production, "bootstrap-class-loader-delegation"));

        // identification
        component.setIdentification(loadIdentificationXml(XPathAPI
            .selectSingleNode(production, "identification")));

        // component class name
        component.setComponentClassName(XMLUtil.getTextContent(XPathAPI
            .selectSingleNode(production, "component-class-name")));

        // component class path
        component.setComponentClassPath(XMLUtil.getTextContents(XPathAPI
            .selectNodeList(production, "component-class-path/path-element")));

        // bootstrap class name
        component.setBootstrapClassName(XMLUtil.getTextContent(XPathAPI
            .selectSingleNode(production, "bootstrap-class-name")));

        // bootstrap class path
        component.setBootstrapClassPath(XMLUtil.getTextContents(XPathAPI
            .selectNodeList(production, "bootstrap-class-path/path-element")));

        // shared library list
        component.setSharedLibraryList(loadSharedLibraryListXml(XPathAPI
            .selectNodeList(production, "shared-library")));

        // installation extension -
        component.setInstallationDescriptorExtension(loadInstallationExtension(
            doc, production));

        return component;
    }

    /**
     * @param connectionsTree
     * @return
     * @throws TransformerException
     */
    protected List<Connection> loadConnectionsXml(Node connectionsTree)
        throws TransformerException {
        List<Connection> result = new ArrayList<Connection>();
        if (connectionsTree == null) {
            return result;
        }

        // get "connections/connection" Node List
        // NodeList list = connectionsTree.getChildNodes();
        NodeList list = XPathAPI.selectNodeList(connectionsTree, "connection");

        int size = list.getLength();

        if (size == 0) {
            return result;
        }

        for (int i = 0; i < size; i++) {
            Node cnElement = list.item(i);

            // connection object
            Connection newItem = new Connection();

            // consumer
            Node consumer = XPathAPI.selectSingleNode(cnElement, "consumer");
            /*
             * TODO deal with cases where there is only the interface name and
             * no endpoint name and service name
             * newItem.setConsumerInterfaceName(extractXmlAttributeQName(consumer,
             * "interface-name"));
             */

            newItem.setConsumerServiceName(extractXmlAttributeQName(consumer,
                "service-name"));

            newItem.setConsumerEndpointName(XMLUtil.getAttributeValue(consumer,
                "endpoint-name"));

            // provider
            Node provider = XPathAPI.selectSingleNode(cnElement, "provider");

            newItem.setProviderServiceName(extractXmlAttributeQName(provider,
                "service-name"));

            newItem.setProviderEndpointName(XMLUtil.getAttributeValue(provider,
                "endpoint-name"));

            result.add(newItem);
        }
        return result;
    }

    /**
     * @param list
     * @return
     */
    protected List<Consumes> loadConsumesXml(NodeList list) {

        if (list == null) {
            return null;
        }

        int size = list.getLength();

        if (size == 0) {
            return null;
        }

        List<Consumes> result = new ArrayList<Consumes>();

        for (int i = 0; i < size; i++) {
            Node consElement = list.item(i);

            // consumes object
            Consumes newItem = new Consumes();

            newItem.setInterfaceName(extractXmlAttributeQName(consElement,
                "interface-name"));

            newItem.setServiceName(extractXmlAttributeQName(consElement,
                "service-name"));

            newItem.setEndpointName(XMLUtil.getAttributeValue(consElement,
                "endpoint-name"));

            newItem.setLinkType(XMLUtil.getAttributeValue(consElement,
                "link-type"));

            result.add(newItem);
        }

        return result;
    }

    protected Identification loadIdentificationXml(Node identificationTree)
        throws TransformerException {
        Identification identification = new Identification();

        identification.setName(XMLUtil.getTextContent(XPathAPI
            .selectSingleNode(identificationTree, "name")));
        identification.setDescription(XMLUtil.getTextContent(XPathAPI
            .selectSingleNode(identificationTree, "description")));

        return identification;
    }

    /**
     * Search in the 'component' node, a node which name is different from those
     * specified by JBI. This node is considered as the installationExtension
     * fragment.
     * 
     * @param doc
     * @param component
     * @return
     */
    protected DocumentFragment loadInstallationExtension(Document doc,
        Node component) {
        DocumentFragment result = null;
        if (doc != null
            && component != null) {
            // search a node which name is different from those attempted
            // this node is considered as the installationExtension node.
            NodeList nl = component.getChildNodes();
            // scan all nodes of 'component' node
            String nodeName = null;
            for (int i = 0; i < nl.getLength()
                && result == null; i++) {
                nodeName = nl.item(i).getNodeName();
                if (Node.TEXT_NODE != nl.item(i).getNodeType()
                    && !"identification".equals(nodeName)
                    && !"component-class-name".equals(nodeName)
                    && !"component-class-path".equals(nodeName)
                    && !"bootstrap-class-name".equals(nodeName)
                    && !"bootstrap-class-path".equals(nodeName)
                    && !"shared-library".equals(nodeName)) {
                    // an extra node is found, create the docFragment if
                    // required
                    if (result == null) {
                        result = doc.createDocumentFragment();
                    }
                    result.appendChild(nl.item(i));
                }
            }
        }
        return result;
    }

    /**
     * @param list
     * @return
     */
    protected List<Provides> loadProvidesXml(NodeList list) {

        if (list == null) {
            return null;
        }

        int size = list.getLength();

        if (size == 0) {
            return null;
        }

        List<Provides> result = new ArrayList<Provides>();

        for (int i = 0; i < size; i++) {
            Node provElement = list.item(i);

            // provides object
            Provides newItem = new Provides();

            newItem.setInterfaceName(extractXmlAttributeQName(provElement,
                "interface-name"));

            newItem.setServiceName(extractXmlAttributeQName(provElement,
                "service-name"));

            newItem.setEndpointName(XMLUtil.getAttributeValue(provElement,
                "endpoint-name"));

            result.add(newItem);
        }

        return result;
    }

    /**
     * @param element
     * @throws TransformerException
     */
    protected ServiceAssembly loadServiceAssemblyXml(Node production)
        throws TransformerException {
        ServiceAssembly serviceAssembly = new ServiceAssembly();

        // identification
        serviceAssembly.setIdentification(loadIdentificationXml(XPathAPI
            .selectSingleNode(production, "identification")));

        // service unit
        serviceAssembly.setServiceUnits(loadServiceUnitListXml(XPathAPI
            .selectNodeList(production, "service-unit")));

        // connections
        serviceAssembly.setConnections(loadConnectionsXml(XPathAPI
            .selectSingleNode(production, "connections")));

        return serviceAssembly;
    }

    /**
     * @param element
     * @throws TransformerException
     */
    protected Services loadServicesXml(Node production)
        throws TransformerException {

        Services services = new Services();

        // is binding component (relative to service unit)
        String isBindingComponent = XMLUtil.getAttributeValue(production,
            "binding-component");
        if (isBindingComponent != null) {
            services.setBindingComponent(Boolean
                .parseBoolean(isBindingComponent));
        }

        // provides
        services.setProvides(loadProvidesXml(XPathAPI.selectNodeList(
            production, "provides")));

        // consumes
        services.setConsumes(loadConsumesXml(XPathAPI.selectNodeList(
            production, "consumes")));

        return services;
    }

    /**
     * @param list
     * @return
     * @throws TransformerException
     */
    protected List<ServiceUnit> loadServiceUnitListXml(NodeList list)
        throws TransformerException {

        List<ServiceUnit> result = new ArrayList<ServiceUnit>();
        if (list == null) {
            return result;
        }

        int size = list.getLength();

        if (size == 0) {
            return result;
        }

        for (int i = 0; i < size; i++) {
            Node suElement = list.item(i);

            // service unit object
            ServiceUnit newItem = new ServiceUnit();

            // identification
            newItem.setIdentification(loadIdentificationXml(XPathAPI
                .selectSingleNode(suElement, "identification")));

            // target
            newItem.setTargetArtifactsZip(XMLUtil.getTextContent(XPathAPI
                .selectSingleNode(suElement, "target/artifacts-zip")));

            newItem.setTargetComponentName(XMLUtil.getTextContent(XPathAPI
                .selectSingleNode(suElement, "target/component-name")));

            result.add(newItem);
        }

        return result;
    }

    /**
     * @param list
     * @return
     */
    protected List<SharedLibraryList> loadSharedLibraryListXml(NodeList list) {
        if (list == null) {
            return null;
        }

        int size = list.getLength();

        if (size == 0) {
            return null;
        }

        List<SharedLibraryList> result = new ArrayList<SharedLibraryList>();

        for (int i = 0; i < size; i++) {
            Node sllElement = list.item(i);

            // shared library list object
            SharedLibraryList newItem = new SharedLibraryList();

            newItem.setName(XMLUtil.getTextContent(sllElement));
            newItem
                .setVersion(XMLUtil.getAttributeValue(sllElement, "version"));

            result.add(newItem);
        }
        return result;
    }

    /**
     * @param element
     * @throws TransformerException
     */
    protected SharedLibrary loadSharedLibraryXml(Node production)
        throws TransformerException {
        SharedLibrary sharedLibrary = new SharedLibrary();

        // shared library attributes
        sharedLibrary.setClassLoaderDelegation(XMLUtil.getAttributeValue(
            production, "class-loader-delegation"));
        sharedLibrary.setVersion(XMLUtil.getAttributeValue(production,
            "version"));

        // identification
        sharedLibrary.setIdentification(loadIdentificationXml(XPathAPI
            .selectSingleNode(production, "identification")));

        // shared library class path
        sharedLibrary.setSharedLibraryClassPath(XMLUtil
            .getTextContents(XPathAPI.selectNodeList(production,
                "shared-library-class-path/path-element")));
        return sharedLibrary;
    }

    /**
     * NOTE: The first location to search for JBI schema files is within the
     * Petals installation root directory. If a schema file is not found within
     * root directory, the alternative is use the packaged schema files flatten
     * somewhere which can be found in class path. It won't work if the schema
     * are compressed in a Zip or Jar archive.
     * 
     * @return
     */
    protected InputStream openJbiSchemaFile() {
        InputStream theJbiSchemaResource = null;

        // find through class path
        // obtain class loader
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        if (cl == null) {
            cl = JBIDescriptorBuilder.class.getClassLoader();
        }

        // load jbi.xsd schema resource file
        URL schemaURL = cl.getResource("jbi.xsd");
        if (schemaURL != null) {
            try {
                theJbiSchemaResource = schemaURL.openStream();
            } catch (IOException e) {
                log.log(Level.SEVERE, e.getMessage(), e);
            }
        }

        if (theJbiSchemaResource == null) {
            // find in petals installation root
            String petalsRoot = System.getProperty("petals.home");
            if (petalsRoot != null) {
                try {
                    theJbiSchemaResource = new FileInputStream(new File(
                        petalsRoot
                            + JBI_SCHEMA_RESOURCE));
                } catch (FileNotFoundException e) {
                    log.log(Level.SEVERE, e.getMessage(), e);
                }
            }
        }

        return theJbiSchemaResource;
    }

}
